#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import dataclasses
from abc import ABCMeta, abstractmethod
from contextlib import asynccontextmanager
from typing import (
    Any,
    AsyncGenerator,
    ClassVar,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)
from uuid import UUID

from argilla_server.enums import (
    MetadataPropertyType,
    RecordSortField,
    ResponseStatus,
    ResponseStatusFilter,
    SimilarityOrder,
    SortOrder,
)
from argilla_server.models import Dataset, MetadataProperty, Record, Response, Suggestion, User, Vector, VectorSettings
from argilla_server.pydantic_v1 import BaseModel, Field, root_validator
from argilla_server.pydantic_v1.generics import GenericModel

__all__ = [
    "SearchEngine",
    "TextQuery",
    "MetadataFilter",
    "TermsMetadataFilter",
    "IntegerMetadataFilter",
    "FloatMetadataFilter",
    "UserResponseStatusFilter",
    "SearchResponseItem",
    "SearchResponses",
    "SortBy",
    "MetadataMetrics",
    "TermsMetadataMetrics",
    "IntegerMetadataMetrics",
    "FloatMetadataMetrics",
    "SuggestionFilterScope",
    "ResponseFilterScope",
    "MetadataFilterScope",
    "RecordFilterScope",
    "FilterScope",
    "TermsFilter",
    "RangeFilter",
    "AndFilter",
    "Filter",
    "Order",
]


@dataclasses.dataclass
class SuggestionFilterScope:
    question: str
    property: str


@dataclasses.dataclass
class ResponseFilterScope:
    question: Optional[str] = None
    property: Optional[str] = None
    user: Optional[User] = None


@dataclasses.dataclass
class MetadataFilterScope:
    metadata_property: str


@dataclasses.dataclass
class RecordFilterScope:
    property: str


FilterScope = Union[SuggestionFilterScope, ResponseFilterScope, MetadataFilterScope, RecordFilterScope]


@dataclasses.dataclass
class TermsFilter:
    scope: FilterScope
    values: List[str]


@dataclasses.dataclass
class RangeFilter:
    scope: FilterScope
    ge: Optional[float] = None
    le: Optional[float] = None


@dataclasses.dataclass
class AndFilter:
    filters: List["Filter"]


Filter = Union[AndFilter, TermsFilter, RangeFilter]


@dataclasses.dataclass
class Order:
    scope: FilterScope
    order: SortOrder


class TextQuery(BaseModel):
    q: str
    field: Optional[str] = None


class UserResponseStatusFilter(BaseModel):
    statuses: List[ResponseStatusFilter]
    user: Optional[User] = None

    class Config:
        arbitrary_types_allowed = True

    @property
    def response_statuses(self) -> List[ResponseStatus]:
        return [
            status.value
            for status in self.statuses
            if status not in [ResponseStatusFilter.pending, ResponseStatusFilter.missing]
        ]

    @property
    def has_pending_status(self) -> bool:
        return ResponseStatusFilter.pending in self.statuses or ResponseStatusFilter.missing in self.statuses


class MetadataFilter(BaseModel):
    metadata_property: MetadataProperty

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    @abstractmethod
    def from_string(cls, metadata_property: MetadataProperty, string: str) -> "MetadataFilter":
        pass


class TermsMetadataFilter(MetadataFilter):
    values: List[str]

    @classmethod
    def from_string(cls, metadata_property: MetadataProperty, string: str) -> "MetadataFilter":
        return cls(metadata_property=metadata_property, values=string.split(","))


NT = TypeVar("NT", int, float)


class _RangeModel(GenericModel, Generic[NT]):
    ge: Optional[NT]
    le: Optional[NT]


class NumericMetadataFilter(GenericModel, Generic[NT], MetadataFilter):
    ge: Optional[NT] = None
    le: Optional[NT] = None

    _json_model: ClassVar[Type[_RangeModel]]

    @root_validator(skip_on_failure=True)
    def check_bounds(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        ge = values.get("ge")
        le = values.get("le")

        if ge is None and le is None:
            raise ValueError("One of 'ge' or 'le' values must be specified")

        if ge is not None and le is not None and ge > le:
            raise ValueError(f"'ge' ({ge}) must be lower or equal than 'le' ({le})")

        return values

    @classmethod
    def from_string(cls, metadata_property: MetadataProperty, string: str) -> "NumericMetadataFilter":
        model = cls._json_model.parse_raw(string)
        return cls(metadata_property=metadata_property, ge=model.ge, le=model.le)


class IntegerMetadataFilter(NumericMetadataFilter[int]):
    _json_model = _RangeModel[int]


class FloatMetadataFilter(NumericMetadataFilter[float]):
    _json_model = _RangeModel[float]


class SearchResponseItem(BaseModel):
    record_id: UUID
    score: Optional[float]


class SearchResponses(BaseModel):
    items: List[SearchResponseItem]
    total: int = 0


class SortBy(BaseModel):
    field: Union[MetadataProperty, RecordSortField]
    order: SortOrder = SortOrder.asc

    class Config:
        arbitrary_types_allowed = True


class TermsMetadataMetrics(BaseModel):
    class TermCount(BaseModel):
        term: str
        count: int

    type: MetadataPropertyType = Field(MetadataPropertyType.terms)
    total: int
    values: List[TermCount] = Field(default_factory=list)


class NumericMetadataMetrics(GenericModel, Generic[NT]):
    min: Optional[NT]
    max: Optional[NT]


class IntegerMetadataMetrics(NumericMetadataMetrics[int]):
    type: MetadataPropertyType = Field(MetadataPropertyType.integer)


class FloatMetadataMetrics(NumericMetadataMetrics[float]):
    type: MetadataPropertyType = Field(MetadataPropertyType.float)


MetadataMetrics = Union[TermsMetadataMetrics, IntegerMetadataMetrics, FloatMetadataMetrics]


class SearchEngine(metaclass=ABCMeta):
    registered_classes = {}

    @classmethod
    @abstractmethod
    async def new_instance(cls) -> "SearchEngine":
        pass

    @abstractmethod
    async def close(self):
        pass

    @classmethod
    def register(cls, engine_name: str):
        def decorator(engine_class):
            cls.registered_classes[engine_name] = engine_class
            return engine_class

        return decorator

    @classmethod
    @asynccontextmanager
    async def get_by_name(cls, engine_name: str) -> AsyncGenerator["SearchEngine", None]:
        engine_name = engine_name.lower().strip()

        if engine_name not in cls.registered_classes:
            raise ValueError(f"No engine class registered for '{engine_name}'")

        engine_class = cls.registered_classes[engine_name]

        engine = None

        try:
            engine = await engine_class.new_instance()
            yield engine
        except Exception as e:
            raise e
        finally:
            if engine is not None:
                await engine.close()

    @abstractmethod
    async def create_index(self, dataset: Dataset):
        pass

    @abstractmethod
    async def delete_index(self, dataset: Dataset):
        pass

    @abstractmethod
    async def configure_metadata_property(self, dataset: Dataset, metadata_property: MetadataProperty):
        pass

    @abstractmethod
    async def index_records(self, dataset: Dataset, records: Iterable[Record]):
        pass

    @abstractmethod
    async def delete_records(self, dataset: Dataset, records: Iterable[Record]):
        pass

    @abstractmethod
    async def update_record_response(self, response: Response):
        pass

    @abstractmethod
    async def delete_record_response(self, response: Response):
        pass

    @abstractmethod
    async def update_record_suggestion(self, suggestion: Suggestion):
        pass

    @abstractmethod
    async def delete_record_suggestion(self, suggestion: Suggestion):
        pass

    @abstractmethod
    async def search(
        self,
        dataset: Dataset,
        query: Optional[Union[TextQuery, str]] = None,
        filter: Optional[Filter] = None,
        sort: Optional[List[Order]] = None,
        # TODO: remove them and keep filter and order
        user_response_status_filter: Optional[UserResponseStatusFilter] = None,
        metadata_filters: Optional[List[MetadataFilter]] = None,
        sort_by: Optional[List[SortBy]] = None,
        # END TODO
        offset: int = 0,
        limit: int = 100,
    ) -> SearchResponses:
        pass

    @abstractmethod
    async def compute_metrics_for(self, metadata_property: MetadataProperty) -> MetadataMetrics:
        pass

    async def configure_index_vectors(self, vector_settings: VectorSettings):
        pass

    @abstractmethod
    async def set_records_vectors(self, dataset: Dataset, vectors: Iterable[Vector]):
        pass

    @abstractmethod
    async def similarity_search(
        self,
        dataset: Dataset,
        vector_settings: VectorSettings,
        value: Optional[List[float]] = None,
        record: Optional[Record] = None,
        query: Optional[Union[TextQuery, str]] = None,
        filter: Optional[Filter] = None,
        # TODO: remove them and keep filter
        user_response_status_filter: Optional[UserResponseStatusFilter] = None,
        metadata_filters: Optional[List[MetadataFilter]] = None,
        # END TODO
        max_results: int = 100,
        order: SimilarityOrder = SimilarityOrder.most_similar,
        threshold: Optional[float] = None,
    ) -> SearchResponses:
        pass



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import TYPE_CHECKING

import pytest
from argilla_server.apis.v0.models.text2text import (
    Text2TextBulkRequest,
    Text2TextRecord,
)
from argilla_server.apis.v0.models.text_classification import (
    TextClassificationBulkRequest,
    TextClassificationRecord,
)
from argilla_server.apis.v0.models.token_classification import (
    TokenClassificationBulkRequest,
    TokenClassificationRecord,
)
from argilla_server.commons.models import TaskType
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import User, UserRole
from argilla_server.services.metrics.models import CommonTasksMetrics
from argilla_server.services.tasks.text_classification.metrics import TextClassificationMetrics
from argilla_server.services.tasks.token_classification.metrics import TokenClassificationMetrics

from tests.factories import UserFactory, WorkspaceFactory

if TYPE_CHECKING:
    from httpx import AsyncClient

COMMON_METRICS_LENGTH = len(CommonTasksMetrics.metrics)
CLASSIFICATION_METRICS_LENGTH = len(TextClassificationMetrics.metrics)


@pytest.mark.asyncio
async def test_wrong_dataset_metrics(async_client: "AsyncClient"):
    workspace = await WorkspaceFactory.create()
    user = await UserFactory.create(role=UserRole.owner, workspaces=[workspace])

    async_client.headers.update({API_KEY_HEADER_NAME: user.api_key})
    workspace_query_params = {"workspace": workspace.name}

    records = [Text2TextRecord.parse_obj(data) for data in [{"text": "this is a text"}] * 4]
    request = Text2TextBulkRequest(records=records)
    dataset = "test_wrong_dataset_metrics"

    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200
    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "task": TaskType.text2text.value, "workspace": workspace.name},
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/Text2Text:bulk",
        json=request.dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.get(
        f"/api/datasets/TokenClassification/{dataset}/metrics", params=workspace_query_params
    )
    assert response.status_code == 400
    assert response.json() == {
        "detail": {
            "code": "argilla.api.errors::WrongTaskError",
            "params": {"message": "Provided task TokenClassification cannot be applied to dataset"},
        }
    }

    response = await async_client.post(
        f"/api/datasets/TokenClassification/{dataset}/metrics/a-metric:summary",
        json={},
        params=workspace_query_params,
    )

    assert response.status_code == 400
    assert response.json() == {
        "detail": {
            "code": "argilla.api.errors::WrongTaskError",
            "params": {"message": "Provided task TokenClassification cannot be applied to dataset"},
        }
    }


@pytest.mark.asyncio
async def test_dataset_for_text2text(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    records = [Text2TextRecord.parse_obj(data) for data in [{"text": "this is a text"}] * 4]
    request = Text2TextBulkRequest(records=records)
    dataset = "test_dataset_for_text2text"

    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200
    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "task": TaskType.text2text.value, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200
    response = await async_client.post(
        f"/api/datasets/{dataset}/Text2Text:bulk",
        json=request.dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.get(
        f"/api/datasets/Text2Text/{dataset}/metrics",
        params=workspace_query_params,
    )
    assert response.status_code == 200

    metrics = response.json()
    assert len(metrics) == COMMON_METRICS_LENGTH


@pytest.mark.asyncio
async def test_dataset_for_token_classification(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    text = "This is a contaminated text"
    metadata = {"metadata": {"field": 1}}
    prediction = {"prediction": {"entities": [], "agent": "test", "score": 0.3}}
    records = [
        TokenClassificationRecord.parse_obj(data)
        for data in [
            {"text": text, "tokens": text.split(" "), **metadata, **prediction},
            {"text": text, "tokens": text.split(" "), **metadata, **prediction},
            {"text": text, "tokens": text.split(" "), **metadata, **prediction},
            {"text": text, "tokens": text.split(" "), **metadata, **prediction},
        ]
    ]

    request = TokenClassificationBulkRequest(records=records)
    dataset = "test_dataset_for_token_classification"

    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200
    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "task": TaskType.token_classification.value, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200
    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=request.dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.get(
        f"/api/datasets/TokenClassification/{dataset}/metrics", params=workspace_query_params
    )
    assert response.status_code == 200

    metrics = response.json()
    assert len(metrics) == len(TokenClassificationMetrics.metrics)

    for metric in metrics:
        metric_id = metric["id"]

        response = await async_client.post(
            f"/api/datasets/TokenClassification/{dataset}/metrics/{metric_id}:summary",
            json={},
            params=workspace_query_params,
        )

        assert response.status_code == 200, f"{metric} :: {response.json()}"
        summary = response.json()

        if "length" not in metric_id and "predicted" not in metric_id and "annotated" not in metric_id:
            assert summary, (metric_id, summary)


@pytest.mark.asyncio
async def test_dataset_metrics(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    records = [
        TextClassificationRecord.parse_obj(data)
        for data in [
            {
                "id": 0,
                "inputs": {"text": "Some test data"},
                "multi_label": False,
                "metadata": {"textLength": len("Some test data")},
            },
            {
                "id": 1,
                "inputs": {"text": "Another data with different length"},
                "multi_label": False,
                "metadata": {"textLength": len("Another data with different length")},
            },
        ]
    ]
    request = TextClassificationBulkRequest(records=records)
    dataset = "test_get_dataset_metrics"

    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification.value, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=request.dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.get(
        f"/api/datasets/TextClassification/{dataset}/metrics", params=workspace_query_params
    )
    assert response.status_code == 200

    metrics = response.json()
    assert len(metrics) == CLASSIFICATION_METRICS_LENGTH

    response = await async_client.post(
        f"/api/datasets/TextClassification/{dataset}/metrics/missing_metric:summary",
        json={},
        params=workspace_query_params,
    )

    assert response.status_code == 404
    assert response.json() == {
        "detail": {
            "code": "argilla.api.errors::EntityNotFoundError",
            "params": {"name": "missing_metric", "type": "ServiceBaseMetric"},
        }
    }

    for metric in metrics:
        response = await async_client.post(
            f"/api/datasets/TextClassification/{dataset}/metrics/{metric['id']}:summary",
            json={},
            params=workspace_query_params,
        )
        assert response.status_code == 200, f"{metric}: {response.json()}"


async def create_some_classification_data(
    async_client: "AsyncClient", dataset: str, workspace_name: str, records: list
):
    workspace_query_params = {"workspace": workspace_name}

    request = TextClassificationBulkRequest(records=[TextClassificationRecord.parse_obj(r) for r in records])

    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        "/api/datasets", json={"name": dataset, "task": TaskType.text_classification.value, "workspace": workspace_name}
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=request.dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200


@pytest.mark.asyncio
async def test_labeling_rule_metric(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_labeling_rule_metric"
    await create_some_classification_data(
        async_client,
        dataset,
        records=[{"inputs": {"text": "This is classification record"}}] * 10,
        workspace_name=argilla_user.username,
    )

    rule_query = "t*"
    response = await async_client.post(
        f"/api/datasets/TextClassification/{dataset}/metrics/labeling_rule:summary?rule_query={rule_query}",
        json={},
        params=workspace_query_params,
    )
    assert response.json() == {
        "annotated_covered_records": 0,
        "correct_records": 0,
        "covered_records": 10,
        "incorrect_records": 0,
    }


@pytest.mark.asyncio
async def test_dataset_labels_for_text_classification(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    records = [
        TextClassificationRecord.parse_obj(data)
        for data in [
            {
                "id": 0,
                "inputs": {"text": "Some test data"},
                "prediction": {"agent": "test", "labels": [{"class": "A"}]},
            },
            {
                "id": 1,
                "inputs": {"text": "Some test data"},
                "annotation": {"agent": "test", "labels": [{"class": "B"}]},
            },
            {
                "id": 2,
                "inputs": {"text": "Some test data"},
                "prediction": {
                    "agent": "test",
                    "labels": [
                        {"class": "A", "score": 0.5},
                        {
                            "class": "D",
                            "score": 0.5,
                        },
                    ],
                },
                "annotation": {"agent": "test", "labels": [{"class": "E"}]},
            },
        ]
    ]
    request = TextClassificationBulkRequest(records=records)
    dataset = "test_dataset_labels_for_text_classification"

    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification.value, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=request.dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/TextClassification/{dataset}/metrics/dataset_labels:summary",
        json={},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = response.json()
    labels = response["labels"]
    assert sorted(labels) == ["A", "B", "D", "E"]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from argilla_server.settings import Settings

from tests.pydantic_v1 import ValidationError


@pytest.mark.parametrize("bad_namespace", ["Badns", "bad-ns", "12-bad-ns", "@bad"])
def test_wrong_settings_namespace(monkeypatch, bad_namespace):
    monkeypatch.setenv("ARGILLA_NAMESPACE", bad_namespace)
    with pytest.raises(ValidationError):
        Settings()


def test_settings_namespace(monkeypatch):
    monkeypatch.setenv("ARGILLA_NAMESPACE", "namespace")
    settings = Settings()

    assert settings.namespace == "namespace"
    assert settings.dataset_index_name == "namespace.ar.datasets"
    assert settings.dataset_records_index_name == "namespace.ar.dataset.{}"


def test_settings_index_replicas_with_shards_defined(monkeypatch):
    monkeypatch.setenv("ARGILLA_ES_RECORDS_INDEX_SHARDS", "100")
    monkeypatch.setenv("ARGILLA_ES_RECORDS_INDEX_REPLICAS", "2")

    settings = Settings()
    assert settings.es_records_index_replicas == 2


def test_settings_default_index_replicas_with_shards_defined(monkeypatch):
    monkeypatch.setenv("ARGILLA_ES_RECORDS_INDEX_SHARDS", "100")
    settings = Settings()

    assert settings.es_records_index_shards == 100
    assert settings.es_records_index_replicas == 0


def test_settings_default_database_url(monkeypatch):
    monkeypatch.setenv("ARGILLA_DATABASE_URL", "")

    settings = Settings()
    assert settings.database_url == f"sqlite+aiosqlite:///{settings.home_path}/argilla.db?check_same_thread=False"


@pytest.mark.parametrize(
    "url, expected_url",
    [
        ("sqlite:///test.db", "sqlite+aiosqlite:///test.db"),
        ("sqlite:///:memory:", "sqlite+aiosqlite:///:memory:"),
        ("postgresql://user:pass@localhost:5432/db", "postgresql+asyncpg://user:pass@localhost:5432/db"),
        ("postgresql+psycopg2://user:pass@localhost:5432/db", "postgresql+asyncpg://user:pass@localhost:5432/db"),
    ],
)
def test_settings_database_url(url: str, expected_url: str, monkeypatch):
    monkeypatch.setenv("ARGILLA_DATABASE_URL", url)
    settings = Settings()
    assert settings.database_url == expected_url



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING

import pytest
from argilla_server.models import User, UserRole
from click.testing import CliRunner
from typer import Typer

from tests.factories import UserSyncFactory

if TYPE_CHECKING:
    from sqlalchemy.orm import Session


@pytest.mark.parametrize("new_role_string", ["owner", "admin"])
def test_update(sync_db: "Session", cli_runner: CliRunner, cli: Typer, new_role_string: str):
    user = UserSyncFactory.create(username="username", role=UserRole.annotator)

    result = cli_runner.invoke(cli, f"database users update username --role {new_role_string}")

    assert result.exit_code == 0

    # user = (await db.execute(select(User).filter_by(username="username"))).scalar_one_or_none()
    user = sync_db.query(User).filter_by(username="username").first()
    assert user.role.value == UserRole(new_role_string).value


def test_update_with_invalid_role(cli_runner: CliRunner, cli: Typer):
    bad_role_str = "bad-role"
    result = cli_runner.invoke(cli, f"database users update username --role {bad_role_str}")

    assert result.exit_code == 2
    assert f"{bad_role_str!r} is not one of" in result.output, result.output


def test_update_with_missing_username(cli_runner: CliRunner, cli: Typer):
    missing_username = "missing-username"
    result = cli_runner.invoke(cli, f"database users update {missing_username} --role owner")

    assert result.exit_code == 0
    assert result.output == f"User with username {missing_username!r} does not exists in database. Skipping...\n"


@pytest.mark.parametrize("role_string", ["owner", "admin", "annotator"])
def test_update_with_same_user_role(cli_runner: CliRunner, cli: Typer, role_string: str):
    username = "username"
    UserSyncFactory.create(username=username, role=UserRole(role_string))

    result = cli_runner.invoke(cli, f"database users update {username} --role {role_string}")

    assert result.exit_code == 0
    assert result.output == f"User {username!r} already has role {role_string!r}. Skipping...\n"



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, Optional, Tuple

from fastapi import Request
from fastapi.security import HTTPBearer
from starlette.authentication import AuthCredentials, AuthenticationBackend, BaseUser

from argilla_server.security.authentication.jwt import JWT
from argilla_server.security.authentication.oauth2.client_provider import OAuth2ClientProvider
from argilla_server.security.authentication.userinfo import UserInfo


class OAuth2AuthenticationBackend(AuthenticationBackend):
    """Authentication backend for AuthenticationMiddleware."""

    scheme = HTTPBearer(auto_error=False)

    def __init__(self, providers: Dict[str, OAuth2ClientProvider]) -> None:
        self.providers = providers

    async def authenticate(self, request: Request) -> Optional[Tuple[AuthCredentials, BaseUser]]:
        credentials = await self.scheme(request)
        if credentials is None:
            return None

        token_data = JWT.decode(credentials.credentials)
        user = UserInfo(token_data)

        provider = self.providers.get(user.get("provider"))
        claims = provider.claims if provider else {}

        return AuthCredentials(user.pop("scope", [])), user.use_claims(claims)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional, Union

from fastapi import APIRouter, Depends, Query, Security

from argilla_server.apis.v0.models.commons.params import CommonTaskHandlerDependencies
from argilla_server.apis.v0.models.text2text import Text2TextQuery, Text2TextRecord
from argilla_server.apis.v0.models.text_classification import TextClassificationQuery, TextClassificationRecord
from argilla_server.apis.v0.models.token_classification import TokenClassificationQuery, TokenClassificationRecord
from argilla_server.commons.config import TasksFactory
from argilla_server.models import User
from argilla_server.pydantic_v1 import BaseModel
from argilla_server.security import auth
from argilla_server.services.datasets import DatasetsService
from argilla_server.services.search.service import SearchRecordsService
from argilla_server.services.storage.service import RecordsStorageService


def configure_router(router: APIRouter):
    QueryType = Union[TextClassificationQuery, TokenClassificationQuery, Text2TextQuery]

    class DeleteRecordsResponse(BaseModel):
        matched: int
        processed: int

    RecordType = Union[
        TextClassificationRecord,
        TokenClassificationRecord,
        Text2TextRecord,
    ]

    @router.get("/{name}/records/{id}", response_model=RecordType)
    async def get_dataset_record(
        name: str,
        id: str,
        request_deps: CommonTaskHandlerDependencies = Depends(),
        service: DatasetsService = Depends(DatasetsService.get_instance),
        search: SearchRecordsService = Depends(SearchRecordsService.get_instance),
        current_user: User = Security(auth.get_current_user),
    ) -> RecordType:
        found = await service.find_by_name(
            user=current_user,
            name=name,
            workspace=request_deps.workspace,
        )

        record_class = TasksFactory.get_task_record(found.task)

        return await search.find_record_by_id(
            dataset=found,
            id=id,
            record_type=record_class,
        )

    @router.delete(
        "/{name}/data",
        operation_id="delete_dataset_records",
        response_model=DeleteRecordsResponse,
    )
    async def delete_dataset_records(
        name: str,
        query: Optional[QueryType] = None,
        mark_as_discarded: bool = Query(
            default=False,
            title="If True, matched records won't be deleted."
            " Instead of that, the record status will be changed to `Discarded`",
        ),
        request_deps: CommonTaskHandlerDependencies = Depends(),
        service: DatasetsService = Depends(DatasetsService.get_instance),
        storage: RecordsStorageService = Depends(RecordsStorageService.get_instance),
        current_user: User = Security(auth.get_current_user),
    ):
        found = await service.find_by_name(
            user=current_user,
            name=name,
            workspace=request_deps.workspace,
        )

        result = await storage.delete_records(
            user=current_user,
            dataset=found,
            query=query,
            mark_as_discarded=mark_as_discarded,
        )

        return DeleteRecordsResponse(
            matched=result.processed,
            processed=result.deleted or result.discarded,
        )


router = APIRouter(tags=["datasets"], prefix="/datasets")
configure_router(router)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
from typing import List

import yaml

from argilla_server.security.authentication.oauth2.client_provider import OAuth2ClientProvider
from argilla_server.security.authentication.oauth2.supported_providers import ALL_SUPPORTED_OAUTH2_PROVIDERS

__all__ = ["OAuth2Settings"]


class AllowedWorkspace:
    def __init__(self, name: str):
        self.name = name


class OAuth2Settings:
    """
    OAuth2 settings model.

    Args:
        enabled:
            Whether OAuth2 authentication is enabled or not.
        allow_http_redirect:
            Whether to allow HTTP scheme on redirect urls (for tests purposes).
        providers:
            List of OAuth2 providers.
        allowed_workspaces:
            List of allowed workspace names (workspace must be created before).
    """

    ALLOWED_WORKSPACES_KEY = "allowed_workspaces"
    PROVIDERS_KEY = "providers"

    def __init__(
        self,
        enabled: bool = True,
        allow_http_redirect: bool = False,
        providers: List[OAuth2ClientProvider] = None,
        allowed_workspaces: List[AllowedWorkspace] = None,
    ):
        self.enabled = enabled
        self.allow_http_redirect = allow_http_redirect
        self.allowed_workspaces = allowed_workspaces or []
        self._providers = providers or []

        if self.allow_http_redirect:
            # See https://stackoverflow.com/questions/27785375/testing-flask-oauthlib-locally-without-https
            os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"

    @property
    def providers(self) -> dict:
        return {provider.name: provider for provider in self._providers}

    @classmethod
    def from_yaml(cls, yaml_file: str) -> "OAuth2Settings":
        """Creates an instance of OAuth2Settings from a YAML file."""

        with open(yaml_file) as f:
            return cls.from_dict(yaml.safe_load(f))

    @classmethod
    def from_dict(cls, settings: dict) -> "OAuth2Settings":
        """Creates an instance of OAuth2Settings from a dictionary."""

        settings[cls.PROVIDERS_KEY] = cls._build_providers(settings)
        settings[cls.ALLOWED_WORKSPACES_KEY] = cls._build_workspaces(settings)

        return cls(**settings)

    @classmethod
    def _build_workspaces(cls, settings: dict) -> List[AllowedWorkspace]:
        return [AllowedWorkspace(**workspace) for workspace in settings.pop(cls.ALLOWED_WORKSPACES_KEY, [])]

    @classmethod
    def _build_providers(cls, settings: dict) -> List["OAuth2ClientProvider"]:
        providers = []

        for provider in settings.pop("providers", []):
            name = provider.pop("name")

            provider_class = ALL_SUPPORTED_OAUTH2_PROVIDERS.get(name)
            if not provider_class:
                raise ValueError(
                    f"Unsupported provider {name}. Supported providers are {ALL_SUPPORTED_OAUTH2_PROVIDERS.keys()}"
                )

            providers.append(provider_class.from_dict(provider))

        return providers



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import datetime
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Type

from fastapi import Depends

from argilla_server.daos.backend import GenericElasticEngineBackend
from argilla_server.daos.backend.base import ClosedIndexError, IndexNotFoundError
from argilla_server.daos.backend.generic_elastic import PaginatedSortInfo
from argilla_server.daos.backend.search.model import BaseRecordsQuery, SortableField
from argilla_server.daos.models.datasets import DatasetDB
from argilla_server.daos.models.records import DaoRecordsSearch, DaoRecordsSearchResults, RecordDB
from argilla_server.errors import ClosedDatasetError, MissingDatasetRecordsError


class DatasetRecordsDAO:
    """Datasets records DAO"""

    _INSTANCE = None

    @classmethod
    def get_instance(
        cls,
        es: GenericElasticEngineBackend = Depends(GenericElasticEngineBackend.get_instance),
    ) -> "DatasetRecordsDAO":
        """
        Creates a dataset records dao instance

        Parameters
        ----------
        es:
            The elasticserach wrapper dependency

        """
        if not cls._INSTANCE:
            cls._INSTANCE = cls(es)
        return cls._INSTANCE

    def __init__(self, es: GenericElasticEngineBackend):
        self._es = es

    def init(self):
        """Initializes dataset records dao. Used on app startup"""
        pass

    def add_records(
        self,
        dataset: DatasetDB,
        records: List[RecordDB],
        record_class: Type[RecordDB],
    ) -> int:
        """
        Add records to dataset

        Parameters
        ----------
        dataset:
            The dataset
        records:
            The list of records
        record_class:
            Record class used to convert records to
        Returns
        -------
            The number of failed records

        """

        now = datetime.datetime.utcnow()
        documents = []
        metadata_values = {}

        mapping = self._es.get_schema(id=dataset.id)
        exclude_fields = [
            name for name in record_class.schema()["properties"] if name not in mapping["mappings"]["properties"]
        ]

        vectors_configuration = {}
        for record in records:
            metadata_values.update(record.metadata or {})

            db_record = record_class.parse_obj(record)
            db_record.last_updated = now

            record_dict = db_record.dict(exclude_none=True, exclude=set(exclude_fields))

            if record.vectors is not None:
                # TODO: Create embeddings config by settings
                for (
                    vector_name,
                    vector_data_mapping,
                ) in record.vectors.items():
                    vector_dimension = vectors_configuration.get(vector_name, None)

                    if vector_dimension is None:
                        dimension = len(vector_data_mapping.value)
                        vectors_configuration[vector_name] = dimension

            documents.append(record_dict)

        self._es.create_dataset(
            id=dataset.id,
            task=dataset.task,
            metadata_values=metadata_values,
            vectors_cfg=vectors_configuration,
        )

        return self._es.add_dataset_records(
            id=dataset.id,
            documents=documents,
        )

    def compute_metric(
        self,
        dataset: DatasetDB,
        metric_id: str,
        metric_params: Dict[str, Any] = None,
        query: Optional[BaseRecordsQuery] = None,
    ):
        return self._es.compute_metric(
            id=dataset.id,
            metric_id=metric_id,
            query=query,
            params=metric_params,
        )

    def search_records(
        self,
        dataset: DatasetDB,
        search: Optional[DaoRecordsSearch] = None,
        size: int = 100,
        record_from: int = 0,
        exclude_fields: List[str] = None,
        highligth_results: bool = True,
    ) -> DaoRecordsSearchResults:
        try:
            search = search or DaoRecordsSearch()

            total, records = self._es.search_records(
                id=dataset.id,
                query=search.query,
                sort=search.sort,
                record_from=record_from,
                size=size,
                exclude_fields=exclude_fields,
                enable_highlight=highligth_results,
            )
            if isinstance(total, dict):
                total = total["value"]
            return DaoRecordsSearchResults(
                total=total,
                records=records,
            )
        except ClosedIndexError:
            raise ClosedDatasetError(dataset.name)
        except IndexNotFoundError:
            raise MissingDatasetRecordsError(f"No records index found for dataset {dataset.name}")

    def scan_dataset(
        self,
        dataset: DatasetDB,
        search: Optional[DaoRecordsSearch] = None,
        limit: Optional[int] = 1000,
        id_from: Optional[str] = None,
        include_fields: Optional[Set[str]] = None,
        exclude_fields: Optional[Set[str]] = None,
    ) -> Iterable[Dict[str, Any]]:
        """
        Iterates over a dataset records

        Parameters
        ----------
        dataset:
            The dataset
        search:
            The search parameters. Optional
        limit:
            Batch size to extract, only works if an `id_from` is provided
        id_from:
            From which ID should we start iterating
        include_fields:
            A set of record fields to retrieve. Wildcard are allowed
        exclude_fields:
            A set of record fields to exclude. Wildcard are allowed

        Returns
        -------
            An iterable over found dataset records
        """

        search = search or DaoRecordsSearch()
        next_search_params = [id_from] if id_from else None
        paginated_sort = PaginatedSortInfo(
            sort_by=[SortableField(id="id")], shuffle=search.sort.shuffle, next_search_params=next_search_params
        )

        return self._es.scan_records(
            id=dataset.id,
            query=search.query,
            sort=paginated_sort,
            limit=limit,
            include_fields=list(include_fields) if include_fields else None,
            exclude_fields=list(exclude_fields) if exclude_fields else None,
        )

    async def delete_records_by_query(
        self,
        dataset: DatasetDB,
        query: Optional[BaseRecordsQuery] = None,
    ) -> Tuple[int, int]:
        total, deleted = await self._es.delete_records_by_query(
            id=dataset.id,
            query=query,
        )
        return total, deleted

    async def update_records_by_query(
        self,
        dataset: DatasetDB,
        query: Optional[BaseRecordsQuery] = None,
        **content,
    ) -> Tuple[int, int]:
        total, updated = await self._es.update_records_content(
            id=dataset.id,
            content=content,
            query=query,
        )
        return total, updated

    async def update_record(
        self,
        dataset: DatasetDB,
        record: RecordDB,
    ):
        self._es.update_record(
            dataset_id=dataset.id,
            record_id=record.id,
            content=record.dict(exclude_none=True),
        )

    async def get_record_by_id(
        self,
        dataset: DatasetDB,
        id: str,
    ) -> Optional[Dict[str, Any]]:
        return self._es.find_record_by_id(
            dataset_id=dataset.id,
            record_id=id,
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import TYPE_CHECKING, Any, Type
from uuid import uuid4

import pytest
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import DatasetStatus, Response, ResponseStatus, UserRole
from argilla_server.search_engine import SearchEngine
from sqlalchemy import func, select

from tests.factories import (
    AdminFactory,
    AnnotatorFactory,
    DatasetFactory,
    LabelSelectionQuestionFactory,
    MultiLabelSelectionQuestionFactory,
    RankingQuestionFactory,
    RatingQuestionFactory,
    RecordFactory,
    ResponseFactory,
    TextQuestionFactory,
    UserFactory,
    WorkspaceFactory,
)

if TYPE_CHECKING:
    from argilla_server.models import User
    from httpx import AsyncClient
    from sqlalchemy.ext.asyncio import AsyncSession

    from tests.factories import QuestionFactory


@pytest.mark.asyncio
class TestSuiteResponses:
    @pytest.mark.parametrize(
        "response_json",
        [
            {
                "values": {"input_ok": {"value": "yes"}, "output_ok": {"value": "yes"}},
                "status": "submitted",
            },
            {
                "values": {"input_ok": {"value": "yes"}},
                "status": "submitted",
            },
            {
                "values": {"output_ok": {"value": "yes"}},
                "status": "draft",
            },
        ],
    )
    async def test_update_response(
        self,
        async_client: "AsyncClient",
        db: "AsyncSession",
        mock_search_engine: SearchEngine,
        owner_auth_header: dict,
        response_json: dict,
    ):
        dataset = await DatasetFactory.create(status=DatasetStatus.ready)
        await TextQuestionFactory.create(name="input_ok", dataset=dataset, required=True)
        await TextQuestionFactory.create(name="output_ok", dataset=dataset)
        record = await RecordFactory.create(dataset=dataset)

        response = await ResponseFactory.create(
            record=record,
            values={"input_ok": {"value": "no"}, "output_ok": {"value": "no"}},
            status=ResponseStatus.submitted,
        )

        dataset_previous_last_activity_at = dataset.last_activity_at
        dataset_previous_updated_at = dataset.updated_at

        resp = await async_client.put(f"/api/v1/responses/{response.id}", headers=owner_auth_header, json=response_json)

        assert resp.status_code == 200
        assert (await db.get(Response, response.id)).values == response_json["values"]

        resp_body = resp.json()
        assert resp_body == {
            "id": str(response.id),
            "values": response_json["values"],
            "status": response_json["status"],
            "record_id": str(response.record_id),
            "user_id": str(response.user_id),
            "inserted_at": response.inserted_at.isoformat(),
            "updated_at": datetime.fromisoformat(resp_body["updated_at"]).isoformat(),
        }

        assert dataset.last_activity_at > dataset_previous_last_activity_at
        assert dataset.updated_at == dataset_previous_updated_at

        mock_search_engine.update_record_response.assert_called_once_with(response)

    async def test_update_response_without_authentication(self, async_client: "AsyncClient", db: "AsyncSession"):
        response = await ResponseFactory.create(
            values={
                "input_ok": {"value": "no"},
                "output_ok": {"value": "no"},
            },
            status="submitted",
        )
        response_json = {
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": "submitted",
        }

        resp = await async_client.put(f"/api/v1/responses/{response.id}", json=response_json)

        assert resp.status_code == 401
        assert (await db.get(Response, response.id)).values == {
            "input_ok": {"value": "no"},
            "output_ok": {"value": "no"},
        }

    @pytest.mark.parametrize(
        "QuestionFactory, response_value",
        [
            (TextQuestionFactory, "Unit Test!"),
            (RatingQuestionFactory, 3),
            (LabelSelectionQuestionFactory, "option1"),
            (MultiLabelSelectionQuestionFactory, ["option1", "option2"]),
            (
                RankingQuestionFactory,
                [
                    {"value": "completion-a", "rank": 1},
                    {"value": "completion-b", "rank": 2},
                    {"value": "completion-c", "rank": 3},
                ],
            ),
        ],
    )
    async def test_update_response_to_submitted_status(
        self,
        async_client: "AsyncClient",
        QuestionFactory: Type["QuestionFactory"],
        response_value: ResponseStatus,
        owner: "User",
        owner_auth_header: dict,
    ):
        question = await QuestionFactory.create()
        record = await RecordFactory.create(dataset=question.dataset)
        response = await ResponseFactory.create(record=record, user=owner, status=ResponseStatus.draft)

        response_json = {"values": {question.name: {"value": response_value}}, "status": ResponseStatus.submitted}
        resp = await async_client.put(f"/api/v1/responses/{response.id}", headers=owner_auth_header, json=response_json)

        assert resp.status_code == 200
        assert resp.json() == {
            "id": str(response.id),
            "record_id": str(record.id),
            "values": {question.name: {"value": response_value}},
            "status": ResponseStatus.submitted.value,
            "user_id": str(owner.id),
            "inserted_at": response.inserted_at.isoformat(),
            "updated_at": response.updated_at.isoformat(),
        }

    @pytest.mark.parametrize(
        "QuestionFactory, response_value",
        [
            (TextQuestionFactory, "Unit Test!"),
            (RatingQuestionFactory, 3),
            (LabelSelectionQuestionFactory, "option1"),
            (MultiLabelSelectionQuestionFactory, ["option1", "option2"]),
            (
                RankingQuestionFactory,
                [
                    {"value": "completion-a", "rank": 1},
                    {"value": "completion-b", "rank": 2},
                    {"value": "completion-c", "rank": 3},
                ],
            ),
            (
                RankingQuestionFactory,
                [
                    {"value": "completion-a", "rank": 1},
                ],
            ),
            (
                RankingQuestionFactory,
                [
                    {"value": "completion-a", "rank": 1},
                    {"value": "completion-b"},
                    {"value": "completion-c"},
                ],
            ),
        ],
    )
    @pytest.mark.parametrize("response_status", [ResponseStatus.draft, ResponseStatus.discarded])
    async def test_update_response_to_non_submitted_status(
        self,
        async_client: "AsyncClient",
        QuestionFactory: Type["QuestionFactory"],
        response_value: Any,
        response_status: ResponseStatus,
        owner: "User",
        owner_auth_header: dict,
    ):
        question = await QuestionFactory.create()
        record = await RecordFactory.create(dataset=question.dataset)
        response = await ResponseFactory.create(record=record, user=owner, status=ResponseStatus.submitted)

        response_json = {"values": {question.name: {"value": response_value}}, "status": response_status}
        resp = await async_client.put(f"/api/v1/responses/{response.id}", headers=owner_auth_header, json=response_json)

        assert resp.status_code == 200
        assert resp.json() == {
            "id": str(response.id),
            "record_id": str(record.id),
            "values": {question.name: {"value": response_value}},
            "status": response_status.value,
            "user_id": str(owner.id),
            "inserted_at": response.inserted_at.isoformat(),
            "updated_at": response.updated_at.isoformat(),
        }

    @pytest.mark.parametrize(
        "QuestionFactory, response_value",
        [
            (TextQuestionFactory, False),
            (RatingQuestionFactory, "wrong-rating-value"),
            (LabelSelectionQuestionFactory, False),
            (MultiLabelSelectionQuestionFactory, "wrong-type"),
            (MultiLabelSelectionQuestionFactory, ["option4", "option5"]),
            (MultiLabelSelectionQuestionFactory, []),
            (RankingQuestionFactory, "wrong-type"),
            (RankingQuestionFactory, []),
            (RankingQuestionFactory, [{"value": "completion-b", "rank": 1}]),
            (
                RankingQuestionFactory,
                [
                    {"value": "completion-b", "rank": 1},
                    {"value": "completion-c", "rank": 2},
                    {"value": "completion-a", "rank": 3},
                    {"value": "completion-z", "rank": 4},
                ],
            ),
            (
                RankingQuestionFactory,
                [
                    {"value": "completion-b", "rank": 1},
                    {"value": "completion-c", "rank": 2},
                    {"value": "completion-a", "rank": 4},
                ],
            ),
            (
                RankingQuestionFactory,
                [
                    {"value": "completion-z", "rank": 1},
                    {"value": "completion-c", "rank": 2},
                    {"value": "completion-a", "rank": 3},
                ],
            ),
            (
                RankingQuestionFactory,
                [
                    {"value": "completion-a", "rank": 1},
                    {"value": "completion-c", "rank": 2},
                    {"value": "completion-a", "rank": 3},
                ],
            ),
        ],
    )
    async def test_update_response_with_wrong_response_value(
        self,
        async_client: "AsyncClient",
        QuestionFactory: Type["QuestionFactory"],
        response_value: Any,
        owner: "User",
        owner_auth_header: dict,
    ):
        question = await QuestionFactory.create()
        record = await RecordFactory.create(dataset=question.dataset)
        response = await ResponseFactory.create(record=record, user=owner, status=ResponseStatus.draft)

        response_json = {"values": {question.name: {"value": response_value}}, "status": ResponseStatus.submitted}
        resp = await async_client.put(f"/api/v1/responses/{response.id}", headers=owner_auth_header, json=response_json)

        assert resp.status_code == 422

    async def test_update_response_as_annotator(self, async_client: "AsyncClient", db: "AsyncSession"):
        dataset = await DatasetFactory.create(status=DatasetStatus.ready)
        await TextQuestionFactory.create(name="input_ok", dataset=dataset)
        await TextQuestionFactory.create(name="output_ok", dataset=dataset)
        record = await RecordFactory.create(dataset=dataset)
        annotator = await AnnotatorFactory.create()

        response = await ResponseFactory.create(
            record=record,
            values={
                "input_ok": {"value": "no"},
                "output_ok": {"value": "no"},
            },
            status="submitted",
            user=annotator,
        )
        response_json = {
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": "submitted",
        }

        resp = await async_client.put(
            f"/api/v1/responses/{response.id}", headers={API_KEY_HEADER_NAME: annotator.api_key}, json=response_json
        )

        assert resp.status_code == 200
        assert (await db.get(Response, response.id)).values == {
            "input_ok": {"value": "yes"},
            "output_ok": {"value": "yes"},
        }

        resp_body = resp.json()
        assert resp_body == {
            "id": str(response.id),
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": "submitted",
            "record_id": str(response.record_id),
            "user_id": str(response.user_id),
            "inserted_at": response.inserted_at.isoformat(),
            "updated_at": datetime.fromisoformat(resp_body["updated_at"]).isoformat(),
        }

    async def test_update_response_as_annotator_for_different_user_response(
        self, async_client: "AsyncClient", db: "AsyncSession"
    ):
        annotator = await AnnotatorFactory.create()
        response = await ResponseFactory.create(
            values={
                "input_ok": {"value": "no"},
                "output_ok": {"value": "no"},
            },
            status="submitted",
        )
        response_json = {
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": "submitted",
        }

        resp = await async_client.put(
            f"/api/v1/responses/{response.id}", headers={API_KEY_HEADER_NAME: annotator.api_key}, json=response_json
        )

        assert resp.status_code == 403
        assert (await db.get(Response, response.id)).values == {
            "input_ok": {"value": "no"},
            "output_ok": {"value": "no"},
        }

    async def test_update_response_with_nonexistent_response_id(
        self, async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
    ):
        response = await ResponseFactory.create(
            values={
                "input_ok": {"value": "no"},
                "output_ok": {"value": "no"},
            },
            status="submitted",
        )
        response_json = {
            "values": {
                "input_ok": {"value": "yes"},
                "output_ok": {"value": "yes"},
            },
            "status": "submitted",
        }

        resp = await async_client.put(f"/api/v1/responses/{uuid4()}", headers=owner_auth_header, json=response_json)

        assert resp.status_code == 404
        assert (await db.get(Response, response.id)).values == {
            "input_ok": {"value": "no"},
            "output_ok": {"value": "no"},
        }

    async def test_delete_response(
        self, async_client: "AsyncClient", mock_search_engine: SearchEngine, db: "AsyncSession", owner_auth_header: dict
    ):
        response = await ResponseFactory.create()
        dataset = response.record.dataset

        dataset_previous_last_activity_at = dataset.last_activity_at
        dataset_previous_updated_at = dataset.updated_at

        resp = await async_client.delete(f"/api/v1/responses/{response.id}", headers=owner_auth_header)

        assert resp.status_code == 200
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 0

        assert dataset.last_activity_at > dataset_previous_last_activity_at
        assert dataset.updated_at == dataset_previous_updated_at

        mock_search_engine.delete_record_response.assert_called_once_with(response)

    async def test_delete_response_without_authentication(self, async_client: "AsyncClient", db: "AsyncSession"):
        response = await ResponseFactory.create()

        resp = await async_client.delete(f"/api/v1/responses/{response.id}")

        assert resp.status_code == 401
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 1

    @pytest.mark.parametrize("role", [UserRole.admin, UserRole.annotator])
    async def test_delete_response_as_restricted_user(
        self, async_client: "AsyncClient", db: "AsyncSession", role: UserRole
    ):
        user = await UserFactory.create(role=role)
        response = await ResponseFactory.create(user=user)

        resp = await async_client.delete(
            f"/api/v1/responses/{response.id}", headers={API_KEY_HEADER_NAME: user.api_key}
        )

        assert resp.status_code == 200
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 0

    async def test_delete_response_as_admin_for_different_user_response(
        self, async_client: "AsyncClient", db: "AsyncSession"
    ):
        workspace = await WorkspaceFactory.create()
        admin = await AdminFactory.create(workspaces=[workspace])
        dataset = await DatasetFactory.create(workspace=workspace)
        record = await RecordFactory.create(dataset=dataset)
        response = await ResponseFactory.create(record=record)

        resp = await async_client.delete(
            f"/api/v1/responses/{response.id}", headers={API_KEY_HEADER_NAME: admin.api_key}
        )

        assert resp.status_code == 200
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 0

    async def test_delete_response_as_annotator_for_different_user_response(
        self, async_client: "AsyncClient", db: "AsyncSession"
    ):
        annotator = await AnnotatorFactory.create()
        response = await ResponseFactory.create()

        resp = await async_client.delete(
            f"/api/v1/responses/{response.id}", headers={API_KEY_HEADER_NAME: annotator.api_key}
        )

        assert resp.status_code == 403
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 1

    async def test_delete_response_with_nonexistent_response_id(
        self, async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
    ):
        await ResponseFactory.create()

        resp = await async_client.delete(f"/api/v1/responses/{uuid4()}", headers=owner_auth_header)

        assert resp.status_code == 404
        assert (await db.execute(select(func.count(Response.id)))).scalar() == 1



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import asyncio
import os
from typing import TYPE_CHECKING, List, Optional

import typer
import yaml

from argilla_server.database import AsyncSessionLocal
from argilla_server.models import User, UserRole
from argilla_server.pydantic_v1 import BaseModel, Field, constr
from argilla_server.schemas.v0.users import USER_USERNAME_REGEX
from argilla_server.schemas.v0.workspaces import WORKSPACE_NAME_REGEX

from .utils import get_or_new_workspace

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession


class WorkspaceCreate(BaseModel):
    name: str = Field(..., regex=WORKSPACE_NAME_REGEX, min_length=1)


class UserCreate(BaseModel):
    first_name: constr(strip_whitespace=True)
    username: str = Field(..., regex=USER_USERNAME_REGEX, min_length=1)
    role: UserRole
    api_key: constr(min_length=1)
    password_hash: constr(min_length=1)
    workspaces: Optional[List[WorkspaceCreate]]


class UsersMigrator:
    def __init__(self, users_filename: str):
        self._users_filename = users_filename

        with open(users_filename) as users_file:
            self._users = yaml.safe_load(users_file.read())

    async def migrate(self):
        typer.echo(f"Starting users migration process using file {self._users_filename!r}")

        async with AsyncSessionLocal() as session:
            try:
                for user in self._users:
                    await self._migrate_user(session, user)
                await session.commit()
            except Exception:
                await session.rollback()
                typer.echo("Users migration process failed...")
                raise typer.Exit(code=1)

            typer.echo("Users migration process successfully finished")

    async def _migrate_user(self, session: "AsyncSession", user: dict):
        typer.echo(f"Migrating User with username {user['username']!r}")

        user_create = self._build_user_create(user)

        await User.create(
            session,
            first_name=user_create.first_name,
            username=user_create.username,
            role=user_create.role,
            api_key=user_create.api_key,
            password_hash=user_create.password_hash,
            workspaces=[await get_or_new_workspace(session, workspace.name) for workspace in user_create.workspaces],
            autocommit=False,
        )

    def _build_user_create(self, user: dict) -> UserCreate:
        return UserCreate(
            first_name=user.get("full_name", ""),
            username=user["username"],
            role=self._user_role(user),
            api_key=user["api_key"],
            password_hash=user["hashed_password"],
            workspaces=[WorkspaceCreate(name=workspace_name) for workspace_name in self._user_workspace_names(user)],
        )

    def _user_role(self, user: dict) -> UserRole:
        if user.get("workspaces") is None:
            return UserRole.owner

        return UserRole.annotator

    def _user_workspace_names(self, user: dict) -> List[str]:
        workspace_names = [workspace_name for workspace_name in user.get("workspaces", [])]

        if user["username"] in workspace_names:
            return workspace_names

        return [user["username"]] + workspace_names


def migrate():
    """Migrate users defined in YAML file to database."""

    users_db_file: str = os.getenv("ARGILLA_LOCAL_AUTH_USERS_DB_FILE", ".users.yml")
    asyncio.run(UsersMigrator(users_db_file).migrate())


if __name__ == "__main__":
    typer.run(migrate)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Iterable, List, Optional

from fastapi import Depends

from argilla_server.commons.config import TasksFactory
from argilla_server.daos.backend.search.model import SortableField
from argilla_server.services.datasets import ServiceBaseDataset
from argilla_server.services.search.model import ServiceSearchResults, ServiceSortConfig
from argilla_server.services.search.service import SearchRecordsService
from argilla_server.services.storage.service import RecordsStorageService
from argilla_server.services.tasks.commons import BulkResponse
from argilla_server.services.tasks.token_classification.model import (
    ServiceTokenClassificationQuery,
    ServiceTokenClassificationRecord,
)


class TokenClassificationService:
    """
    Token classification service

    """

    _INSTANCE: Optional["TokenClassificationService"] = None

    @classmethod
    def get_instance(
        cls,
        storage: RecordsStorageService = Depends(RecordsStorageService.get_instance),
        search: SearchRecordsService = Depends(SearchRecordsService.get_instance),
    ):
        if not cls._INSTANCE:
            cls._INSTANCE = cls(storage, search)
        return cls._INSTANCE

    def __init__(
        self,
        storage: RecordsStorageService,
        search: SearchRecordsService,
    ):
        self.__storage__ = storage
        self.__search__ = search

    async def add_records(
        self,
        dataset: ServiceBaseDataset,
        records: List[ServiceTokenClassificationRecord],
    ):
        failed = await self.__storage__.store_records(
            dataset=dataset,
            records=records,
            record_type=ServiceTokenClassificationRecord,
        )
        return BulkResponse(dataset=dataset.name, processed=len(records), failed=failed)

    def search(
        self,
        dataset: ServiceBaseDataset,
        query: ServiceTokenClassificationQuery,
        sort_by: List[SortableField],
        record_from: int = 0,
        size: int = 100,
        exclude_metrics: bool = True,
    ) -> ServiceSearchResults:
        """
        Run a search in a dataset

        Parameters
        ----------
        dataset:
            The records dataset
        query:
            The search parameters
        sort_by:
            The sort by order list
        record_from:
            The record from return results
        size:
            The max number of records to return

        Returns
        -------
            The matched records with aggregation info for specified task_meta.py

        """
        metrics = TasksFactory.find_task_metrics(
            dataset.task,
            metric_ids={
                "words_cloud",
                "predicted_by",
                "predicted_as",
                "annotated_by",
                "annotated_as",
                "error_distribution",
                "predicted_mentions_distribution",
                "annotated_mentions_distribution",
                "status_distribution",
                "metadata",
                "score",
            },
        )

        results = self.__search__.search(
            dataset,
            query=query,
            record_type=ServiceTokenClassificationRecord,
            size=size,
            metrics=metrics,
            record_from=record_from,
            exclude_metrics=exclude_metrics,
            sort_config=ServiceSortConfig(sort_by=sort_by),
        )

        if results.metrics:
            results.metrics["words"] = results.metrics.get("words_cloud", {})
            results.metrics["status"] = results.metrics["status_distribution"]
            results.metrics["predicted"] = results.metrics["error_distribution"]
            results.metrics["predicted"].pop("unknown", None)
            results.metrics["mentions"] = results.metrics["annotated_mentions_distribution"]
            results.metrics["predicted_mentions"] = results.metrics["predicted_mentions_distribution"]

        return results

    def read_dataset(
        self,
        dataset: ServiceBaseDataset,
        query: ServiceTokenClassificationQuery,
        id_from: Optional[str] = None,
        limit: int = 1000,
    ) -> Iterable[ServiceTokenClassificationRecord]:
        """
        Scan a dataset records

        Parameters
        ----------
        dataset:
            The dataset name
        query:
            If provided, scan will retrieve only records matching
            the provided query filters. Optional
        id_from:
            If provided, read the samples after this record ID
        limit:
            Batch size to scan, only used if `id_from` is specified

        """
        yield from self.__search__.scan_records(
            dataset,
            query=query,
            record_type=ServiceTokenClassificationRecord,
            id_from=id_from,
            limit=limit,
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional, Set, Type

from argilla_server.commons.models import TaskType
from argilla_server.errors import EntityNotFoundError, WrongTaskError
from argilla_server.pydantic_v1 import BaseModel
from argilla_server.services.datasets import ServiceBaseDataset, ServiceDataset
from argilla_server.services.metrics import ServiceBaseMetric
from argilla_server.services.metrics.models import ServiceBaseTaskMetrics
from argilla_server.services.search.model import ServiceRecordsQuery
from argilla_server.services.tasks.commons import ServiceRecord


class TaskConfig(BaseModel):
    task: TaskType
    query: Type[ServiceRecordsQuery]
    dataset: Type[ServiceDataset]
    record: Type[ServiceRecord]
    metrics: Optional[Type[ServiceBaseTaskMetrics]]


class TasksFactory:
    __REGISTERED_TASKS__ = dict()

    @classmethod
    def register_task(
        cls,
        task_type: TaskType,
        query_request: Type[ServiceRecordsQuery],
        record_class: Type[ServiceRecord],
        dataset_class: Optional[Type[ServiceDataset]] = None,
        metrics: Optional[Type[ServiceBaseTaskMetrics]] = None,
    ):
        cls.__REGISTERED_TASKS__[task_type] = TaskConfig(
            task=task_type,
            dataset=dataset_class or ServiceBaseDataset,
            query=query_request,
            record=record_class,
            metrics=metrics,
        )

    @classmethod
    def get_all_configs(cls) -> List[TaskConfig]:
        return [cfg for cfg in cls.__REGISTERED_TASKS__.values()]

    @classmethod
    def get_task_by_task_type(cls, task_type: TaskType) -> Optional[TaskConfig]:
        return cls.__REGISTERED_TASKS__.get(task_type)

    @classmethod
    def get_task_metrics(cls, task: TaskType) -> Optional[Type[ServiceBaseTaskMetrics]]:
        config = cls.get_task_by_task_type(task)
        if config:
            return config.metrics

    @classmethod
    def get_task_dataset(cls, task: TaskType) -> Type[ServiceDataset]:
        config = cls.__get_task_config__(task)
        return config.dataset

    @classmethod
    def get_task_record(cls, task: TaskType) -> Type[ServiceRecord]:
        config = cls.__get_task_config__(task)
        return config.record

    @classmethod
    def __get_task_config__(cls, task):
        config = cls.get_task_by_task_type(task)
        if not config:
            raise WrongTaskError(f"No configuration found for task {task}")
        return config

    @classmethod
    def find_task_metric(cls, task: TaskType, metric_id: str) -> Optional[ServiceBaseMetric]:
        metrics = cls.find_task_metrics(task, {metric_id})
        if metrics:
            return metrics[0]
        raise EntityNotFoundError(name=metric_id, type=ServiceBaseMetric)

    @classmethod
    def find_task_metrics(cls, task: TaskType, metric_ids: Set[str]) -> List[ServiceBaseMetric]:
        if not metric_ids:
            return []

        metrics = []
        for metric in cls.get_task_metrics(task).metrics:
            if metric.id in metric_ids:
                metrics.append(metric)
        return metrics



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime, timedelta

from jose import JWTError, jwt

from argilla_server.errors import UnauthorizedError
from argilla_server.security.authentication.userinfo import UserInfo
from argilla_server.security.settings import settings


class JWT:
    secret: str = settings.secret_key
    algorithm: str = settings.algorithm
    expires: int = settings.token_expiration

    @classmethod
    def encode(cls, data: dict) -> str:
        return jwt.encode(data, cls.secret, algorithm=cls.algorithm)

    @classmethod
    def decode(cls, token: str) -> dict:
        try:
            return jwt.decode(token, cls.secret, algorithms=[cls.algorithm])
        except JWTError:
            raise UnauthorizedError("Invalid token")

    @classmethod
    def create(cls, user: UserInfo) -> str:
        expire = datetime.utcnow() + timedelta(seconds=cls.expires)
        return cls.encode({**user, "exp": expire})



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar, Union
from uuid import UUID

from fastapi import HTTPException, Query

from argilla_server.pydantic_v1 import BaseModel


# TODO: remove this function at some point
def parse_uuids(uuids_str: str) -> List[UUID]:
    try:
        return [UUID(uuid_str) for uuid_str in uuids_str.split(",")]
    except ValueError:
        raise HTTPException(status_code=422, detail="Invalid UUID format")


T = TypeVar("T", bound=BaseModel)

# Matches key1,key2,key3 and key:value1,value2,value3, but not key1,key2,key3:value1,value2,value3
PARAM_REGEX = re.compile(r"^(?:(?:[^,:]+(?:,[^,:]+)*)|(?:[^,:]+:[^,:]+(?:,[^,:]+)*))$")


def parse_query_param(
    name: str,
    max_keys: Optional[int] = None,
    max_values_per_key: Optional[int] = None,
    model: Optional[Type[T]] = None,
    group_keys_without_values: bool = True,
    **kwargs: Any,
) -> Callable[[Optional[List[str]]], Union[Dict[str, Any], T, None]]:
    """Generates a function that can be used as a FastAPI dependency (`fastapi.Depends`) and that parses the values of
    a query parameter with the following format:

        `...?param_name=key1&param_name=key2&param_name=key3,key4&param_name=key5:value1,value2,value3`

    The values are parsed into a dictionary with the following structure:

        ```python
        {
            "keys": ["key1", "key2", "key3", "key4"],
            "key5": ["value1", "value2", "value3"]
        }
        ```

    or, if `max_keys` is set to 1:

        ```python
        {
            "key": "key1",
            "value": ["value1", "value2", "value3"]
        }
        ```

    or, if `group_keys_without_values` is set to `False`:

        ```python
        {
            "key1": None,
            "key2": None,
            "key3": None,
            "key4": None,
            "key5": ["value1", "value2", "value3"]
        }
        ```

    In addition, if a `pydantic.BaseModel` is provided, the dictionary is parsed into an instance of that model:

        ```python
        from argilla_server.pydantic_v1 import BaseModel, Field


        class Params(BaseModel):
            keys: List[str] = Field(..., alias="keys")
            my_other_key_with_values: List[str] = Field(..., alias="key5")
        ```

    Args:
        name: name of the query parameter.
        max_keys: maximum number of keys allowed in the query parameter. Defaults to `None`.
        max_values_per_key: maximum number of values per key allowed in the query parameter. Defaults to `None`.
        model: optional `pydantic.BaseModel` to parse the query parameter into. Defaults to `None`.
        kwargs: extra parameters for `pydantic.Query` function.

    Returns:
        A function that parses the query parameter into a dictionary or a `pydantic.BaseModel` instance.
    """

    # Doing this for the correctness of the OpenAPI schema
    if max_keys is not None and max_keys == 1:
        query_param_typing = Optional[str]
    else:
        query_param_typing = Optional[List[str]]

    def _parse(
        param_values: Optional[query_param_typing] = Query(None, alias=name, **kwargs)
    ) -> Union[Dict[str, Any], T, None]:
        if param_values is None:
            return None

        if isinstance(param_values, str):
            param_values = [param_values]

        parsed_params: Dict[str, Any] = defaultdict(list)
        for value in param_values:
            if not PARAM_REGEX.match(value):
                raise HTTPException(
                    status_code=422,
                    detail=f"'{name}' query parameter must be of the form 'key1,key2,key3' or 'key:value1,value2,value3'",
                )

            parts = value.split(":")
            num_parts = len(parts)
            if num_parts == 1:
                parts = parts[0].split(",")
                if group_keys_without_values:
                    parsed_params["keys"].extend(parts)
                else:
                    for part in parts:
                        parsed_params[part] = None
            else:
                key = parts[0]
                values = parts[1].split(",")
                parsed_params[key].extend(values)

        # Check that the number of keys is within the allowed range
        num_keys_without_values = len(parsed_params.get("keys", []))
        num_keys_with_values = len(parsed_params) - 1  # we don't want to count the "keys" key
        total_keys = num_keys_with_values + num_keys_without_values
        if max_keys is not None and total_keys > max_keys:
            raise HTTPException(
                status_code=422, detail=f"'{name}' query parameter must contain at most {max_keys} comma-separated keys"
            )

        # Check that the number of values per key is within the allowed range
        if max_values_per_key is not None:
            for key, values in parsed_params.items():
                if key == "keys":
                    continue

                num_values = len(values) if values is not None else 0
                if num_values > max_values_per_key:
                    raise HTTPException(
                        status_code=422,
                        detail=(
                            f"'{name}' query parameter must contain at most {max_values_per_key} values per"
                            f" comma-separated key. '{key}' has {num_values} values."
                        ),
                    )

                if num_values == 1:
                    parsed_params[key] = values[0]

        # If `max_keys == 1`, we want to return a dictionary with just `key` and `value` keys
        if max_keys == 1:
            if num_keys_without_values == 1:
                parsed_params["key"] = parsed_params["keys"].pop()
                del parsed_params["keys"]
            else:
                key, value = next(iter(parsed_params.items()))
                parsed_params["key"] = key
                parsed_params["value"] = value
                del parsed_params[key]

        if model is not None:
            return model(**parsed_params)

        return parsed_params

    return _parse



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from httpx import AsyncClient


@pytest.mark.asyncio
@pytest.mark.parametrize("http_method", ["GET", "POST", "PUT", "DELETE", "PATCH"])
async def test_route_not_found_response(async_client: AsyncClient, http_method: str):
    response = await async_client.request(method=http_method, url="/api/not/found/route")

    assert response.status_code == 404
    assert response.json() == {"detail": "Endpoint '/api/not/found/route' not found"}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Any, List, Mapping
from uuid import UUID

from sqlalchemy import func, select
from sqlalchemy.dialects import postgresql
from sqlalchemy.ext.asyncio import AsyncSession

from argilla_server.contexts.datasets import (
    get_metadata_property_by_name_and_dataset_id_or_raise,
    get_question_by_name_and_dataset_id_or_raise,
)
from argilla_server.models import Question, Suggestion
from argilla_server.schemas.v1.records import (
    FilterScope,
    MetadataFilterScope,
    RecordFilterScope,
    SearchRecordsQuery,
)
from argilla_server.schemas.v1.responses import ResponseFilterScope
from argilla_server.schemas.v1.suggestions import SuggestionFilterScope


class SearchRecordsQueryValidator:
    def __init__(self, db: AsyncSession, query: SearchRecordsQuery, dataset_id: UUID):
        self._db = db
        self._query = query
        self._dataset_id = dataset_id

    async def validate(self) -> None:
        if self._query.filters:
            for filter in self._query.filters.and_:
                await self._validate_filter_scope(filter.scope)

        if self._query.sort:
            for order in self._query.sort:
                await self._validate_filter_scope(order.scope)

    async def _validate_filter_scope(self, filter_scope: FilterScope) -> None:
        if isinstance(filter_scope, RecordFilterScope):
            return
        elif isinstance(filter_scope, ResponseFilterScope):
            await self._validate_response_filter_scope(filter_scope)
        elif isinstance(filter_scope, SuggestionFilterScope):
            await self._validate_suggestion_filter_scope(filter_scope)
        elif isinstance(filter_scope, MetadataFilterScope):
            await self._validate_metadata_filter_scope(filter_scope)
        else:
            raise ValueError(f"Unknown filter scope entity `{filter_scope.entity}`")

    async def _validate_response_filter_scope(self, filter_scope: ResponseFilterScope) -> None:
        if filter_scope.question is None:
            return

        await get_question_by_name_and_dataset_id_or_raise(self._db, filter_scope.question, self._dataset_id)

    async def _validate_suggestion_filter_scope(self, filter_scope: SuggestionFilterScope) -> None:
        await get_question_by_name_and_dataset_id_or_raise(self._db, filter_scope.question, self._dataset_id)

    async def _validate_metadata_filter_scope(self, filter_scope: MetadataFilterScope) -> None:
        await get_metadata_property_by_name_and_dataset_id_or_raise(
            self._db, filter_scope.metadata_property, self._dataset_id
        )


async def validate_search_records_query(db: AsyncSession, query: SearchRecordsQuery, dataset_id: UUID) -> None:
    await SearchRecordsQueryValidator(db, query, dataset_id).validate()


async def get_dataset_suggestion_agents_by_question(db: AsyncSession, dataset_id: UUID) -> List[Mapping[str, Any]]:
    if db.bind.dialect.name == postgresql.dialect.name:
        return await _get_dataset_suggestion_agents_by_question_postgresql(db, dataset_id)
    else:
        return await _get_dataset_suggestion_agents_by_question_sqlite(db, dataset_id)


async def _get_dataset_suggestion_agents_by_question_postgresql(
    db: AsyncSession, dataset_id: UUID
) -> List[Mapping[str, Any]]:
    result = await db.execute(
        select(
            Question.id.label("question_id"),
            Question.name.label("question_name"),
            func.array_remove(func.array_agg(Suggestion.agent.distinct()), None).label("suggestion_agents"),
        )
        .outerjoin(Suggestion)
        .where(Question.dataset_id == dataset_id)
        .group_by(Question.id)
        .order_by(Question.inserted_at)
    )

    return result.mappings().all()


async def _get_dataset_suggestion_agents_by_question_sqlite(
    db: AsyncSession, dataset_id: UUID
) -> List[Mapping[str, Any]]:
    result = await db.execute(
        select(
            Question.id.label("question_id"),
            Question.name.label("question_name"),
            func.group_concat(Suggestion.agent.distinct()).label("suggestion_agents"),
        )
        .outerjoin(Suggestion)
        .where(Question.dataset_id == dataset_id)
        .group_by(Question.id)
        .order_by(Question.inserted_at)
    )

    rows = result.mappings().all()

    return [
        {
            "question_id": row["question_id"],
            "question_name": row["question_name"],
            "suggestion_agents": row["suggestion_agents"].split(",") if row["suggestion_agents"] else [],
        }
        for row in rows
    ]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import inspect
import random

import factory
from argilla_server.enums import FieldType, MetadataPropertyType
from argilla_server.models import (
    Dataset,
    Field,
    MetadataProperty,
    Question,
    QuestionType,
    Record,
    Response,
    Suggestion,
    User,
    UserRole,
    Vector,
    VectorSettings,
    Workspace,
    WorkspaceUser,
)
from argilla_server.models.base import DatabaseModel
from factory.alchemy import SESSION_PERSISTENCE_COMMIT, SESSION_PERSISTENCE_FLUSH
from factory.builder import BuildStep, StepBuilder, parse_declarations
from sqlalchemy.ext.asyncio import async_object_session

from tests.database import SyncTestSession, TestSession


# https://github.com/FactoryBoy/factory_boy/issues/679#issuecomment-1348746070
class AsyncSQLAlchemyModelFactory(factory.alchemy.SQLAlchemyModelFactory):
    @classmethod
    async def _generate(cls, strategy, params):
        if cls._meta.abstract:
            raise factory.errors.FactoryError(
                "Cannot generate instances of abstract factory %(f)s; "
                "Ensure %(f)s.Meta.model is set and %(f)s.Meta.abstract "
                "is either not set or False." % dict(f=cls.__name__)
            )

        step = AsyncStepBuilder(cls._meta, params, strategy)
        return await step.build()

    @classmethod
    async def _create(cls, model_class, *args, **kwargs):
        for key, value in kwargs.items():
            if inspect.isawaitable(value):
                kwargs[key] = await value
            if isinstance(value, DatabaseModel):
                old_session = async_object_session(value)
                session = cls._meta.sqlalchemy_session.registry().sync_session
                if old_session.sync_session.hash_key != session.hash_key:
                    old_session.expunge(value)
                    value = await old_session.merge(value)
        return await super()._create(model_class, *args, **kwargs)

    @classmethod
    async def create_batch(cls, size, **kwargs):
        return [await cls.create(**kwargs) for _ in range(size)]

    @classmethod
    async def _save(cls, model_class, session, args, kwargs):
        session_persistence = cls._meta.sqlalchemy_session_persistence
        obj = model_class(*args, **kwargs)
        session.add(obj)
        if session_persistence == SESSION_PERSISTENCE_FLUSH:
            await session.flush()
        elif session_persistence == SESSION_PERSISTENCE_COMMIT:
            await session.commit()
        return obj


class AsyncStepBuilder(StepBuilder):
    # Redefine build function that await for instance creation and awaitable postgenerations
    async def build(self, parent_step=None, force_sequence=None):
        pre, post = parse_declarations(
            self.extras,
            base_pre=self.factory_meta.pre_declarations,
            base_post=self.factory_meta.post_declarations,
        )

        if force_sequence is not None:
            sequence = force_sequence
        elif self.force_init_sequence is not None:
            sequence = self.force_init_sequence
        else:
            sequence = self.factory_meta.next_sequence()

        step = BuildStep(
            builder=self,
            sequence=sequence,
            parent_step=parent_step,
        )
        step.resolve(pre)

        args, kwargs = self.factory_meta.prepare_arguments(step.attributes)

        instance = await self.factory_meta.instantiate(
            step=step,
            args=args,
            kwargs=kwargs,
        )

        postgen_results = {}
        for declaration_name in post.sorted():
            declaration = post[declaration_name]
            declaration_result = declaration.declaration.evaluate_post(
                instance=instance,
                step=step,
                overrides=declaration.context,
            )
            if inspect.isawaitable(declaration_result):
                declaration_result = await declaration_result
            postgen_results[declaration_name] = declaration_result

        self.factory_meta.use_postgeneration_results(
            instance=instance,
            step=step,
            results=postgen_results,
        )
        return instance


class BaseFactory(AsyncSQLAlchemyModelFactory):
    class Meta:
        sqlalchemy_session = TestSession
        sqlalchemy_session_persistence = "flush"


class BaseSyncFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        sqlalchemy_session = SyncTestSession
        sqlalchemy_session_persistence = "flush"


class WorkspaceUserFactory(BaseFactory):
    class Meta:
        model = WorkspaceUser


class WorkspaceFactory(BaseFactory):
    class Meta:
        model = Workspace

    name = factory.Sequence(lambda n: f"workspace-{n}")


class WorkspaceSyncFactory(BaseSyncFactory):
    class Meta:
        model = Workspace

    name = factory.Sequence(lambda n: f"workspace-{n}")


class UserFactory(BaseFactory):
    class Meta:
        model = User

    first_name = factory.Faker("first_name")
    username = factory.Sequence(lambda n: f"username-{n}")
    api_key = factory.Sequence(lambda n: f"api-key-{n}")
    password_hash = "$2y$05$eaw.j2Kaw8s8vpscVIZMfuqSIX3OLmxA21WjtWicDdn0losQ91Hw."


class UserSyncFactory(BaseSyncFactory):
    class Meta:
        model = User

    first_name = factory.Faker("first_name")
    username = factory.Sequence(lambda n: f"username-{n}")
    api_key = factory.Sequence(lambda n: f"api-key-{n}")
    password_hash = "$2y$05$eaw.j2Kaw8s8vpscVIZMfuqSIX3OLmxA21WjtWicDdn0losQ91Hw."


class OwnerFactory(UserFactory):
    role = UserRole.owner


class AdminFactory(UserFactory):
    role = UserRole.admin


class AnnotatorFactory(UserFactory):
    role = UserRole.annotator


class DatasetFactory(BaseFactory):
    class Meta:
        model = Dataset

    name = factory.Sequence(lambda n: f"dataset-{n}")
    workspace = factory.SubFactory(WorkspaceFactory)


class RecordFactory(BaseFactory):
    class Meta:
        model = Record

    fields = {
        "text": "This is a text",
        "sentiment": "neutral",
    }
    external_id = factory.Sequence(lambda n: f"external-id-{n}")
    dataset = factory.SubFactory(DatasetFactory)


class ResponseFactory(BaseFactory):
    class Meta:
        model = Response

    record = factory.SubFactory(RecordFactory)
    user = factory.SubFactory(UserFactory)


class VectorSettingsFactory(BaseFactory):
    class Meta:
        model = VectorSettings

    name = factory.Sequence(lambda n: f"vector-{n}")
    title = "Vector Title"
    dimensions = factory.LazyAttribute(lambda _: random.randrange(16, 1024))
    dataset = factory.SubFactory(DatasetFactory)


class VectorFactory(BaseFactory):
    class Meta:
        model = Vector

    record = factory.SubFactory(RecordFactory)
    vector_settings = factory.SubFactory(VectorSettingsFactory)


class FieldFactory(BaseFactory):
    class Meta:
        model = Field

    name = factory.Sequence(lambda n: f"field-{n}")
    title = "Field Title"
    dataset = factory.SubFactory(DatasetFactory)


class TextFieldFactory(FieldFactory):
    settings = {"type": FieldType.text.value, "use_markdown": False}


class MetadataPropertyFactory(BaseFactory):
    class Meta:
        model = MetadataProperty

    # TODO: Remove this method and fix possible failing tests
    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        default_settings = getattr(cls, "settings", {})
        settings = kwargs.get("settings", {})
        if settings:
            new_settings = default_settings.copy()
            new_settings.update(settings)
            kwargs["settings"] = new_settings
        return super()._create(model_class, *args, **kwargs)

    name = factory.Sequence(lambda n: f"metadata-property-{n}")
    title = "Metadata property title"
    allowed_roles = [UserRole.admin, UserRole.annotator]
    dataset = factory.SubFactory(DatasetFactory)


class TermsMetadataPropertyFactory(MetadataPropertyFactory):
    settings = {"type": MetadataPropertyType.terms, "values": ["a", "b", "c"]}


class IntegerMetadataPropertyFactory(MetadataPropertyFactory):
    settings = {"type": MetadataPropertyType.integer}


class FloatMetadataPropertyFactory(MetadataPropertyFactory):
    settings = {"type": MetadataPropertyType.float}


class QuestionFactory(BaseFactory):
    class Meta:
        model = Question

    # TODO: Remove this method and fix possible failing tests
    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        default_settings = cls.settings.copy()
        settings = kwargs.get("settings", {})
        if settings:
            new_settings = default_settings.copy()
            new_settings.update(settings)
            kwargs["settings"] = new_settings
        return super()._create(model_class, *args, **kwargs)

    name = factory.Sequence(lambda n: f"question-{n}")
    title = "Question Title"
    description = "Question Description"
    dataset = factory.SubFactory(DatasetFactory)
    settings = {}


class TextQuestionFactory(QuestionFactory):
    settings = {"type": QuestionType.text.value, "use_markdown": False}


class RatingQuestionFactory(QuestionFactory):
    settings = {
        "type": QuestionType.rating.value,
        "options": [
            {"value": 1},
            {"value": 2},
            {"value": 3},
            {"value": 4},
            {"value": 5},
            {"value": 6},
            {"value": 7},
            {"value": 8},
            {"value": 9},
            {"value": 10},
        ],
    }


class LabelSelectionQuestionFactory(QuestionFactory):
    settings = {
        "type": QuestionType.label_selection.value,
        "options": [
            {"value": "option1", "text": "Option 1", "description": None},
            {"value": "option2", "text": "Option 2", "description": None},
            {"value": "option3", "text": "Option 3", "description": None},
        ],
    }


class MultiLabelSelectionQuestionFactory(QuestionFactory):
    settings = {
        "type": QuestionType.multi_label_selection.value,
        "options": [
            {"value": "option1", "text": "Option 1", "description": None},
            {"value": "option2", "text": "Option 2", "description": None},
            {"value": "option3", "text": "Option 3", "description": None},
        ],
    }


class RankingQuestionFactory(QuestionFactory):
    settings = {
        "type": QuestionType.ranking.value,
        "options": [
            {"value": "completion-a", "text": "Completion A", "description": None},
            {"value": "completion-b", "text": "Completion B", "description": None},
            {"value": "completion-c", "text": "Completion C", "description": None},
        ],
    }


class SuggestionFactory(BaseFactory):
    class Meta:
        model = Suggestion

    record = factory.SubFactory(RecordFactory)
    question = factory.SubFactory(QuestionFactory)
    value = "negative"



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re
from typing import Dict, List, Optional, Tuple, Union
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Query, Security, status
from sqlalchemy.ext.asyncio import AsyncSession
from typing_extensions import Annotated

import argilla_server.errors.future as errors
import argilla_server.search_engine as search_engine
from argilla_server.apis.v1.handlers.datasets.datasets import _get_dataset
from argilla_server.contexts import datasets, search
from argilla_server.database import get_async_db
from argilla_server.enums import MetadataPropertyType, RecordSortField, ResponseStatusFilter, SortOrder
from argilla_server.models import Dataset as DatasetModel
from argilla_server.models import Record, User
from argilla_server.policies import DatasetPolicyV1, authorize
from argilla_server.schemas.v1.datasets import Dataset
from argilla_server.schemas.v1.records import (
    Filters,
    FilterScope,
    MetadataFilterScope,
    MetadataParsedQueryParam,
    MetadataQueryParams,
    Order,
    RangeFilter,
    RecordFilterScope,
    RecordIncludeParam,
    Records,
    RecordsCreate,
    RecordsUpdate,
    SearchRecord,
    SearchRecordsQuery,
    SearchRecordsResult,
    TermsFilter,
)
from argilla_server.schemas.v1.records import Record as RecordSchema
from argilla_server.schemas.v1.responses import ResponseFilterScope
from argilla_server.schemas.v1.suggestions import (
    SearchSuggestionOptions,
    SearchSuggestionOptionsQuestion,
    SearchSuggestionsOptions,
    SuggestionFilterScope,
)
from argilla_server.schemas.v1.vector_settings import VectorSettings
from argilla_server.search_engine import (
    AndFilter,
    FloatMetadataFilter,
    IntegerMetadataFilter,
    MetadataFilter,
    SearchEngine,
    SearchResponses,
    SortBy,
    TermsMetadataFilter,
    UserResponseStatusFilter,
    get_search_engine,
)
from argilla_server.security import auth
from argilla_server.telemetry import TelemetryClient, get_telemetry_client
from argilla_server.utils import parse_query_param, parse_uuids

LIST_DATASET_RECORDS_LIMIT_DEFAULT = 50
LIST_DATASET_RECORDS_LIMIT_LE = 1000
LIST_DATASET_RECORDS_DEFAULT_SORT_BY = {RecordSortField.inserted_at.value: "asc"}
DELETE_DATASET_RECORDS_LIMIT = 100

_RECORD_SORT_FIELD_VALUES = tuple(field.value for field in RecordSortField)
_VALID_SORT_VALUES = tuple(sort.value for sort in SortOrder)
_METADATA_PROPERTY_SORT_BY_REGEX = re.compile(r"^metadata\.(?P<name>(?=.*[a-z0-9])[a-z0-9_-]+)$")


SortByQueryParamParsed = Annotated[
    Dict[str, str],
    Depends(
        parse_query_param(
            name="sort_by",
            description=(
                "The field used to sort the records. Expected format is `field` or `field:{asc,desc}`, where `field`"
                " can be 'inserted_at', 'updated_at' or the name of a metadata property"
            ),
            max_values_per_key=1,
            group_keys_without_values=False,
        )
    ),
]

parse_record_include_param = parse_query_param(
    name="include", help="Relationships to include in the response", model=RecordIncludeParam
)

router = APIRouter()


async def _filter_records_using_search_engine(
    db: "AsyncSession",
    search_engine: "SearchEngine",
    dataset: Dataset,
    parsed_metadata: List[MetadataParsedQueryParam],
    limit: int,
    offset: int,
    user: Optional[User] = None,
    response_statuses: Optional[List[ResponseStatusFilter]] = None,
    include: Optional[RecordIncludeParam] = None,
    sort_by_query_param: Optional[Dict[str, str]] = None,
) -> Tuple[List["Record"], int]:
    search_responses = await _get_search_responses(
        db=db,
        search_engine=search_engine,
        dataset=dataset,
        limit=limit,
        offset=offset,
        user=user,
        parsed_metadata=parsed_metadata,
        response_statuses=response_statuses,
        sort_by_query_param=sort_by_query_param,
    )

    record_ids = [response.record_id for response in search_responses.items]
    user_id = user.id if user else None

    return (
        await datasets.get_records_by_ids(
            db=db, dataset_id=dataset.id, user_id=user_id, records_ids=record_ids, include=include
        ),
        search_responses.total,
    )


def _to_search_engine_filter_scope(scope: FilterScope, user: Optional[User]) -> search_engine.FilterScope:
    if isinstance(scope, RecordFilterScope):
        return search_engine.RecordFilterScope(property=scope.property)
    elif isinstance(scope, MetadataFilterScope):
        return search_engine.MetadataFilterScope(metadata_property=scope.metadata_property)
    elif isinstance(scope, SuggestionFilterScope):
        return search_engine.SuggestionFilterScope(question=scope.question, property=scope.property)
    elif isinstance(scope, ResponseFilterScope):
        return search_engine.ResponseFilterScope(question=scope.question, property=scope.property, user=user)
    else:
        raise Exception(f"Unknown scope type {type(scope)}")


def _to_search_engine_filter(filters: Filters, user: Optional[User]) -> search_engine.Filter:
    engine_filters = []

    for filter in filters.and_:
        engine_scope = _to_search_engine_filter_scope(filter.scope, user=user)

        if isinstance(filter, TermsFilter):
            engine_filter = search_engine.TermsFilter(scope=engine_scope, values=filter.values)
        elif isinstance(filter, RangeFilter):
            engine_filter = search_engine.RangeFilter(scope=engine_scope, ge=filter.ge, le=filter.le)
        else:
            raise Exception(f"Unknown filter type {type(filter)}")

        engine_filters.append(engine_filter)

    return AndFilter(filters=engine_filters)


def _to_search_engine_sort(sort: List[Order], user: Optional[User]) -> List[search_engine.Order]:
    engine_sort = []

    for order in sort:
        engine_scope = _to_search_engine_filter_scope(order.scope, user=user)
        engine_sort.append(search_engine.Order(scope=engine_scope, order=order.order))

    return engine_sort


async def _get_search_responses(
    db: "AsyncSession",
    search_engine: "SearchEngine",
    dataset: DatasetModel,
    parsed_metadata: List[MetadataParsedQueryParam],
    limit: int,
    offset: int,
    search_records_query: Optional[SearchRecordsQuery] = None,
    user: Optional[User] = None,
    response_statuses: Optional[List[ResponseStatusFilter]] = None,
    sort_by_query_param: Optional[Dict[str, str]] = None,
) -> "SearchResponses":
    search_records_query = search_records_query or SearchRecordsQuery()

    text_query = None
    vector_query = None
    if search_records_query.query:
        text_query = search_records_query.query.text
        vector_query = search_records_query.query.vector

    filters = search_records_query.filters
    sort = search_records_query.sort

    vector_settings = None
    record = None

    if vector_query:
        vector_settings = await _get_vector_settings_by_name_or_raise(db, dataset, vector_query.name)
        if vector_query.record_id is not None:
            record = await _get_dataset_record_by_id_or_raise(db, dataset, vector_query.record_id)
            await record.awaitable_attrs.vectors

            if not record.vector_value_by_vector_settings(vector_settings):
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=f"Record `{record.id}` does not have a vector for vector settings `{vector_settings.name}`",
                )

    if (
        text_query
        and text_query.field
        and not await datasets.get_field_by_name_and_dataset_id(db, text_query.field, dataset.id)
    ):
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Field `{text_query.field}` not found in dataset `{dataset.id}`.",
        )

    metadata_filters = await _build_metadata_filters(db, dataset, parsed_metadata)
    response_status_filter = await _build_response_status_filter_for_search(response_statuses, user=user)
    sort_by = await _build_sort_by(db, dataset, sort_by_query_param)

    if vector_query and vector_settings:
        similarity_search_params = {
            "dataset": dataset,
            "vector_settings": vector_settings,
            "value": vector_query.value,
            "record": record,
            "query": text_query,
            "order": vector_query.order,
            "metadata_filters": metadata_filters,
            "user_response_status_filter": response_status_filter,
            "max_results": limit,
        }

        if filters:
            similarity_search_params["filter"] = _to_search_engine_filter(filters, user=user)

        return await search_engine.similarity_search(**similarity_search_params)
    else:
        search_params = {
            "dataset": dataset,
            "query": text_query,
            "metadata_filters": metadata_filters,
            "user_response_status_filter": response_status_filter,
            "offset": offset,
            "limit": limit,
            "sort_by": sort_by,
        }

        if filters:
            search_params["filter"] = _to_search_engine_filter(filters, user=user)
        if sort:
            search_params["sort"] = _to_search_engine_sort(sort, user=user)

        return await search_engine.search(**search_params)


async def _build_metadata_filters(
    db: "AsyncSession", dataset: Dataset, parsed_metadata: List[MetadataParsedQueryParam]
) -> List["MetadataFilter"]:
    try:
        metadata_filters = []
        for metadata_param in parsed_metadata:
            metadata_property = await datasets.get_metadata_property_by_name_and_dataset_id(
                db, name=metadata_param.name, dataset_id=dataset.id
            )
            if metadata_property is None:
                continue  # won't fail on unknown metadata filter name

            if metadata_property.type == MetadataPropertyType.terms:
                metadata_filter_class = TermsMetadataFilter
            elif metadata_property.type == MetadataPropertyType.integer:
                metadata_filter_class = IntegerMetadataFilter
            elif metadata_property.type == MetadataPropertyType.float:
                metadata_filter_class = FloatMetadataFilter
            else:
                raise ValueError(f"Not found filter for type {metadata_property.type}")

            metadata_filters.append(metadata_filter_class.from_string(metadata_property, metadata_param.value))
    except ValueError as ex:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=f"Cannot parse provided metadata filters: {ex}"
        )
    return metadata_filters


async def _build_response_status_filter_for_search(
    response_statuses: Optional[List[ResponseStatusFilter]] = None, user: Optional[User] = None
) -> Optional[UserResponseStatusFilter]:
    user_response_status_filter = None

    if response_statuses:
        # TODO(@frascuchon): user response and status responses should be split into different filter types
        user_response_status_filter = UserResponseStatusFilter(user=user, statuses=response_statuses)

    return user_response_status_filter


async def _build_sort_by(
    db: "AsyncSession", dataset: Dataset, sort_by_query_param: Optional[Dict[str, str]] = None
) -> Union[List[SortBy], None]:
    if sort_by_query_param is None:
        return None

    sorts_by = []
    for sort_field, sort_order in sort_by_query_param.items():
        if sort_field in _RECORD_SORT_FIELD_VALUES:
            field = sort_field
        elif (match := _METADATA_PROPERTY_SORT_BY_REGEX.match(sort_field)) is not None:
            metadata_property_name = match.group("name")
            metadata_property = await datasets.get_metadata_property_by_name_and_dataset_id(
                db, name=metadata_property_name, dataset_id=dataset.id
            )
            if not metadata_property:
                raise HTTPException(
                    status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=(
                        f"Provided metadata property in 'sort_by' query param '{metadata_property_name}' not found in"
                        f" dataset with '{dataset.id}'."
                    ),
                )
            field = metadata_property
        else:
            valid_sort_fields = ", ".join(f"'{sort_field}'" for sort_field in _RECORD_SORT_FIELD_VALUES)
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=(
                    f"Provided sort field in 'sort_by' query param '{sort_field}' is not valid. It must be either"
                    f" {valid_sort_fields} or `metadata.metadata-property-name`"
                ),
            )

        if sort_order is not None and sort_order not in _VALID_SORT_VALUES:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=(
                    f"Provided sort order in 'sort_by' query param '{sort_order}' for field '{sort_field}' is not"
                    " valid."
                ),
            )

        sorts_by.append(SortBy(field=field, order=sort_order or SortOrder.asc.value))

    return sorts_by


async def _validate_search_records_query(db: "AsyncSession", query: SearchRecordsQuery, dataset_id: UUID):
    try:
        await search.validate_search_records_query(db, query, dataset_id)
    except (ValueError, errors.NotFoundError) as e:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(e))


async def _get_dataset_record_by_id_or_raise(db: "AsyncSession", dataset: Dataset, record_id: UUID) -> "Record":
    record = await datasets.get_record_by_id(db, record_id)
    if record is None or record.dataset_id != dataset.id:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Record with id `{record_id}` not found in dataset `{dataset.id}`.",
        )

    return record


async def _get_vector_settings_by_name_or_raise(
    db: "AsyncSession", dataset: Dataset, vector_name: str
) -> VectorSettings:
    vector_settings = await datasets.get_vector_settings_by_name_and_dataset_id(
        db, name=vector_name, dataset_id=dataset.id
    )

    if vector_settings is None:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Vector `{vector_name}` not found in dataset `{dataset.id}`.",
        )

    return vector_settings


@router.get("/me/datasets/{dataset_id}/records", response_model=Records, response_model_exclude_unset=True)
async def list_current_user_dataset_records(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    dataset_id: UUID,
    metadata: MetadataQueryParams = Depends(),
    sort_by_query_param: SortByQueryParamParsed,
    include: Optional[RecordIncludeParam] = Depends(parse_record_include_param),
    response_statuses: List[ResponseStatusFilter] = Query([], alias="response_status"),
    offset: int = 0,
    limit: int = Query(default=LIST_DATASET_RECORDS_LIMIT_DEFAULT, ge=1, le=LIST_DATASET_RECORDS_LIMIT_LE),
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(db, dataset_id)

    await authorize(current_user, DatasetPolicyV1.get(dataset))

    records, total = await _filter_records_using_search_engine(
        db,
        search_engine,
        dataset=dataset,
        parsed_metadata=metadata.metadata_parsed,
        limit=limit,
        offset=offset,
        user=current_user,
        response_statuses=response_statuses,
        include=include,
        sort_by_query_param=sort_by_query_param or LIST_DATASET_RECORDS_DEFAULT_SORT_BY,
    )

    return Records(items=records, total=total)


@router.get("/datasets/{dataset_id}/records", response_model=Records, response_model_exclude_unset=True)
async def list_dataset_records(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    dataset_id: UUID,
    metadata: MetadataQueryParams = Depends(),
    sort_by_query_param: SortByQueryParamParsed,
    include: Optional[RecordIncludeParam] = Depends(parse_record_include_param),
    response_statuses: List[ResponseStatusFilter] = Query([], alias="response_status"),
    offset: int = 0,
    limit: int = Query(default=LIST_DATASET_RECORDS_LIMIT_DEFAULT, ge=1, le=LIST_DATASET_RECORDS_LIMIT_LE),
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(db, dataset_id)

    await authorize(current_user, DatasetPolicyV1.list_records_with_all_responses(dataset))

    records, total = await _filter_records_using_search_engine(
        db,
        search_engine,
        dataset=dataset,
        parsed_metadata=metadata.metadata_parsed,
        limit=limit,
        offset=offset,
        response_statuses=response_statuses,
        include=include,
        sort_by_query_param=sort_by_query_param or LIST_DATASET_RECORDS_DEFAULT_SORT_BY,
    )

    return Records(items=records, total=total)


@router.post("/datasets/{dataset_id}/records", status_code=status.HTTP_204_NO_CONTENT)
async def create_dataset_records(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    telemetry_client: TelemetryClient = Depends(get_telemetry_client),
    dataset_id: UUID,
    records_create: RecordsCreate,
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(
        db, dataset_id, with_fields=True, with_questions=True, with_metadata_properties=True, with_vectors_settings=True
    )

    await authorize(current_user, DatasetPolicyV1.create_records(dataset))

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    #  After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        await datasets.create_records(db, search_engine, dataset=dataset, records_create=records_create)
        telemetry_client.track_data(action="DatasetRecordsCreated", data={"records": len(records_create.items)})
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.patch("/datasets/{dataset_id}/records", status_code=status.HTTP_204_NO_CONTENT)
async def update_dataset_records(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    telemetry_client: TelemetryClient = Depends(get_telemetry_client),
    dataset_id: UUID,
    records_update: RecordsUpdate,
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(db, dataset_id, with_fields=True, with_questions=True, with_metadata_properties=True)

    await authorize(current_user, DatasetPolicyV1.update_records(dataset))

    try:
        await datasets.update_records(db, search_engine, dataset, records_update)
        telemetry_client.track_data(action="DatasetRecordsUpdated", data={"records": len(records_update.items)})
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.delete("/datasets/{dataset_id}/records", status_code=status.HTTP_204_NO_CONTENT)
async def delete_dataset_records(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    dataset_id: UUID,
    current_user: User = Security(auth.get_current_user),
    ids: str = Query(..., description="A comma separated list with the IDs of the records to be removed"),
):
    dataset = await _get_dataset(db, dataset_id)

    await authorize(current_user, DatasetPolicyV1.delete_records(dataset))

    record_ids = parse_uuids(ids)
    num_records = len(record_ids)

    if num_records == 0:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail="No record IDs provided")

    if num_records > DELETE_DATASET_RECORDS_LIMIT:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Cannot delete more than {DELETE_DATASET_RECORDS_LIMIT} records at once",
        )

    await datasets.delete_records(db, search_engine, dataset, record_ids)


@router.post(
    "/me/datasets/{dataset_id}/records/search",
    status_code=status.HTTP_200_OK,
    response_model=SearchRecordsResult,
    response_model_exclude_unset=True,
)
async def search_current_user_dataset_records(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    telemetry_client: TelemetryClient = Depends(get_telemetry_client),
    dataset_id: UUID,
    body: SearchRecordsQuery,
    metadata: MetadataQueryParams = Depends(),
    sort_by_query_param: SortByQueryParamParsed,
    include: Optional[RecordIncludeParam] = Depends(parse_record_include_param),
    response_statuses: List[ResponseStatusFilter] = Query([], alias="response_status"),
    offset: int = Query(0, ge=0),
    limit: int = Query(default=LIST_DATASET_RECORDS_LIMIT_DEFAULT, ge=1, le=LIST_DATASET_RECORDS_LIMIT_LE),
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(db, dataset_id, with_fields=True)

    await authorize(current_user, DatasetPolicyV1.search_records(dataset))

    await _validate_search_records_query(db, body, dataset_id)

    search_responses = await _get_search_responses(
        db=db,
        search_engine=search_engine,
        dataset=dataset,
        search_records_query=body,
        parsed_metadata=metadata.metadata_parsed,
        limit=limit,
        offset=offset,
        user=current_user,
        response_statuses=response_statuses,
        sort_by_query_param=sort_by_query_param,
    )

    record_id_score_map = {
        response.record_id: {"query_score": response.score, "search_record": None}
        for response in search_responses.items
    }

    records = await datasets.get_records_by_ids(
        db=db,
        dataset_id=dataset_id,
        records_ids=list(record_id_score_map.keys()),
        include=include,
        user_id=current_user.id,
    )

    for record in records:
        record_id_score_map[record.id]["search_record"] = SearchRecord(
            record=RecordSchema.from_orm(record), query_score=record_id_score_map[record.id]["query_score"]
        )

    return SearchRecordsResult(
        items=[record["search_record"] for record in record_id_score_map.values()], total=search_responses.total
    )


@router.post(
    "/datasets/{dataset_id}/records/search",
    status_code=status.HTTP_200_OK,
    response_model=SearchRecordsResult,
    response_model_exclude_unset=True,
)
async def search_dataset_records(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    dataset_id: UUID,
    body: SearchRecordsQuery,
    metadata: MetadataQueryParams = Depends(),
    sort_by_query_param: SortByQueryParamParsed,
    include: Optional[RecordIncludeParam] = Depends(parse_record_include_param),
    response_statuses: List[ResponseStatusFilter] = Query([], alias="response_status"),
    offset: int = Query(0, ge=0),
    limit: int = Query(default=LIST_DATASET_RECORDS_LIMIT_DEFAULT, ge=1, le=LIST_DATASET_RECORDS_LIMIT_LE),
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(db, dataset_id, with_fields=True)

    await authorize(current_user, DatasetPolicyV1.search_records_with_all_responses(dataset))

    await _validate_search_records_query(db, body, dataset_id)

    search_responses = await _get_search_responses(
        db=db,
        search_engine=search_engine,
        dataset=dataset,
        search_records_query=body,
        limit=limit,
        offset=offset,
        parsed_metadata=metadata.metadata_parsed,
        response_statuses=response_statuses,
        sort_by_query_param=sort_by_query_param,
    )

    record_id_score_map = {
        response.record_id: {"query_score": response.score, "search_record": None}
        for response in search_responses.items
    }

    records = await datasets.get_records_by_ids(
        db=db,
        dataset_id=dataset_id,
        records_ids=list(record_id_score_map.keys()),
        include=include,
    )

    for record in records:
        record_id_score_map[record.id]["search_record"] = SearchRecord(
            record=RecordSchema.from_orm(record), query_score=record_id_score_map[record.id]["query_score"]
        )

    return SearchRecordsResult(
        items=[record["search_record"] for record in record_id_score_map.values()], total=search_responses.total
    )


@router.get(
    "/datasets/{dataset_id}/records/search/suggestions/options",
    status_code=status.HTTP_200_OK,
    response_model=SearchSuggestionsOptions,
)
async def list_dataset_records_search_suggestions_options(
    *,
    db: AsyncSession = Depends(get_async_db),
    dataset_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    dataset = await _get_dataset(db, dataset_id)

    await authorize(current_user, DatasetPolicyV1.search_records(dataset))

    suggestion_agents_by_question = await search.get_dataset_suggestion_agents_by_question(db, dataset.id)

    return SearchSuggestionsOptions(
        items=[
            SearchSuggestionOptions(
                question=SearchSuggestionOptionsQuestion(id=sa["question_id"], name=sa["question_name"]),
                agents=sa["suggestion_agents"],
            )
            for sa in suggestion_agents_by_question
        ]
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Set, TypeVar, Union

from sqlalchemy import sql
from sqlalchemy.dialects.mysql import insert as mysql_insert
from sqlalchemy.dialects.postgresql import insert as postgres_insert
from sqlalchemy.dialects.sqlite import insert as sqlite_insert
from sqlalchemy.engine.default import DefaultExecutionContext
from sqlalchemy.orm import Mapped, mapped_column
from typing_extensions import Self

from argilla_server.pydantic_v1 import BaseModel

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession
    from sqlalchemy.orm import InstrumentedAttribute
    from sqlalchemy.sql.elements import BinaryExpression

Schema = TypeVar("Schema", bound=BaseModel)

# TODO: remove this once SQLAlchemy has an agnostic insert function
_INSERT_FUNC = {
    "sqlite": sqlite_insert,
    "postgresql": postgres_insert,
    "mysql": mysql_insert,
}


def _schema_or_kwargs(schema: Union[Schema, None], values: Dict[str, Any]) -> Dict[str, Any]:
    if schema:
        return schema.dict()
    return values


class CRUDMixin:
    __upsertable_columns__: Union[Set[str], None] = None

    def fill(self, replace_dict: bool = False, **kwargs: Any) -> Self:
        for key, value in kwargs.items():
            if not hasattr(self, key):
                raise AttributeError(f"Model `{self.__class__.__name__}` has no attribute `{key}`")
            # If the value is a dict, set value for each key one by one, as we want to update only the keys that are in
            # `value` and not override the whole dict.
            if isinstance(value, dict) and not replace_dict:
                dict_col = getattr(self, key) or {}
                dict_col.update(value)
                value = dict_col
            setattr(self, key, value)
        return self

    @classmethod
    async def create(
        cls, db: "AsyncSession", schema: Union[Schema, None] = None, autocommit: bool = True, **kwargs: Any
    ) -> Self:
        _values = _schema_or_kwargs(schema, kwargs)
        instance = cls()
        instance.fill(**_values)
        return await instance.save(db, autocommit)

    @classmethod
    async def read(cls, db: "AsyncSession", id: Any, key: str = "id") -> Union[Self, None]:
        params = {key: id}
        return await cls.read_by(db, **params)

    @classmethod
    async def read_by(cls, db: "AsyncSession", **params: Any) -> Union[Self, None]:
        query = sql.select(cls).filter_by(**params)
        result = await db.execute(query)
        return result.scalars().unique().one_or_none()

    async def update(
        self,
        db: "AsyncSession",
        schema: Union[Schema, None] = None,
        replace_dict: bool = False,
        autocommit: bool = True,
        **kwargs: Any,
    ) -> Self:
        _values = _schema_or_kwargs(schema, kwargs)
        updated = self.fill(replace_dict=replace_dict, **_values)
        return await updated.save(db, autocommit)

    @classmethod
    async def update_many(
        cls,
        db: "AsyncSession",
        objects: List[Dict[str, Any]],
        autocommit: bool = True,
    ) -> None:
        if len(objects) == 0:
            raise ValueError("Cannot update empty list of objects")

        await db.execute(sql.update(cls), objects)

        if autocommit:
            await db.commit()

    @classmethod
    async def upsert_many(
        cls,
        db: "AsyncSession",
        objects: List[Union[Schema, dict]],
        constraints: List["InstrumentedAttribute[Any]"],
        autocommit: bool = True,
    ) -> List[Self]:
        if len(objects) == 0:
            raise ValueError("Cannot upsert empty list of objects")

        values = [obj if isinstance(obj, dict) else obj.dict() for obj in objects]

        # Try to insert all objects
        insert_stmt = _INSERT_FUNC[db.bind.dialect.name](cls).values(values)

        # On conflict, update the columns that are upsertable (defined in `Model.__upsertable_columns__`)
        columns_to_update = {column: insert_stmt.excluded[column] for column in cls.__upsertable_columns__}

        # onupdate for `updated_at` is not working. We need to force a new value on update
        if hasattr(cls, "updated_at"):
            columns_to_update["updated_at"] = datetime.utcnow()
        upsert_stmt = (
            insert_stmt.on_conflict_do_update(index_elements=constraints, set_=columns_to_update)
            .returning(cls)
            .execution_options(populate_existing=True)
        )

        result = await db.execute(upsert_stmt)
        if autocommit:
            await db.commit()

        return result.scalars().all()

    @classmethod
    async def upsert(
        cls,
        db: "AsyncSession",
        schema: Union[Schema, dict],
        constraints: List["InstrumentedAttribute[Any]"],
        autocommit: bool = True,
    ) -> Self:
        upserted = await cls.upsert_many(db, [schema], constraints, autocommit)

        return upserted[0]

    async def delete(self, db: "AsyncSession", autocommit: bool = True) -> Self:
        await db.delete(self)
        if autocommit:
            await db.commit()
        return self

    @classmethod
    async def delete_many(
        cls, db: "AsyncSession", params: List["BinaryExpression"], autocommit: bool = True
    ) -> List[Self]:
        delete_stmt = sql.delete(cls).filter(*params).returning(cls)
        result = await db.execute(delete_stmt)
        if autocommit:
            await db.commit()
        return result.scalars().all()

    async def save(self, db: "AsyncSession", autocommit: bool = True) -> Self:
        db.add(self)
        if autocommit:
            await db.commit()
        return self


def inserted_at_current_value(context: DefaultExecutionContext) -> datetime:
    return context.get_current_parameters(isolate_multiinsert_groups=False)["inserted_at"]


class TimestampMixin:
    inserted_at: Mapped[datetime] = mapped_column(default=datetime.utcnow)
    updated_at: Mapped[datetime] = mapped_column(default=inserted_at_current_value, onupdate=datetime.utcnow)
